#------------------------------------------------------------------
# $Copyright: Copyright (c) 2013 Symantec Corporation.
# All rights reserved.
#
# THIS SOFTWARE CONTAINS CONFIDENTIAL INFORMATION AND TRADE SECRETS OF
# SYMANTEC CORPORATION.  USE, DISCLOSURE OR REPRODUCTION IS PROHIBITED
# WITHOUT THE PRIOR EXPRESS WRITTEN PERMISSION OF SYMANTEC CORPORATION.
#
# The Licensed Software and Documentation are deemed to be commercial
# computer software as defined in FAR 12.212 and subject to restricted
# rights as defined in FAR Section 52.227-19 "Commercial Computer
# Software - Restricted Rights" and DFARS 227.7202, "Rights in
# Commercial Computer Software or Commercial Computer Software
# Documentation", as applicable, and any successor regulations. Any use,
# modification, reproduction release, performance, display or disclosure
# of the Licensed Software and Documentation by the U.S. Government
# shall be solely in accordance with the terms of this Agreement.  $
#
# $Header: /project/isv-agent/VProAgents/SAPLiveCache/SAPLiveCache.pm,v 1.18 2013/06/05 14:19:58 avaidya Exp $
#
# $Author: avaidya $
# $Date: 2013/06/05 14:19:58 $
# $Revision: 1.18 $
#
# Provides:                            SAPLiveCache
# Author:                              vchappav
# CrAgent Run Date:                    01 Nov, 2006
# Modification History:
# Known bugs/caveats:
#------------------------------------------------------------------

#------------
# Be strict..
#------------
use strict;

#----------------------
# Declare the package..
#----------------------
package SAPLiveCache;

use IPC::SysV qw ( IPC_STAT IPC_RMID );

my $sCmdIpcs   = "/usr/bin/ipcs";

my ( $CLUS_HOME, $sPrefix ) = ();

#---------------------
# Setup VCS/VCSOne home..
#---------------------
if ( $ENV { CLUSTER_HOME } ) { # VCS 5.0/VCS One..
  $CLUS_HOME ||= $ENV { CLUSTER_HOME } ||= "/opt/VRTSvcsone";
}
else { # VCS 4.x..
  $CLUS_HOME ||= $ENV { VCS_HOME } ||= "/opt/VRTSvcs";
}

( $sPrefix ) = $CLUS_HOME =~ /.*(vcs(|one))$/;

if ( $sPrefix =~ /vcsone/ ) {
  $sPrefix = "ha";
}

#-----------------------------------------
# Define module-vendor specific commands..
#-----------------------------------------
my $sCmdHares  = "$CLUS_HOME/bin/hares";
my $sCmdHatype = "$CLUS_HOME/bin/hatype";
my $sCmdVcsDecrypt = "$CLUS_HOME/sbin/${sPrefix}decrypt";

my @liRunningProcs = (); # Global..

#-------------------------------------
# Include required framework modules..
#-------------------------------------
use Sys;            # Support generalized UNIX commands
use Proc;           # For calls to the process routines
use Comms;          # For calls to socket(3C) routines
use Version;        # Version display generator
use VCSagentFW;     # Primary Agent Interface to the ACC Library
use VCSDefines;     # This contains constants that VCS uses

#--------------------------------
# Include standard PERL modules..
#--------------------------------
use Fcntl;
use Socket;
use File::Path qw ( mkpath );
use File::stat;

#-----------------------------------------------------------------------
# Add our version number to the version hash and check if versions are
# requested..
#-----------------------------------------------------------------------
my $sVer = "\$Id: SAPLiveCache.pm,v 1.18 2013/06/05 14:19:58 avaidya Exp $::ETVHeader";
Version::TellVersion ( $sVer );

#--------------------------------
# Define SAPLiveCache subroutines..
#--------------------------------

=pod

 -------------------------------------------------------------------------
 Function :  bool ArgsValid ( IN String EntryPointName,
                              IN String EnvFile,
                              IN String LCHome,
                              IN String LCUser,
                              IN String LCName,
                              IN String DBMUserId,
                              IN String DBMPassword,
                              IN String APOHost,
                              IN String InstId,
                              IN String ClientId,
                              IN String RFCDest,
                              IN String RFCHome,
                              IN String RFCUserId,
                              IN String RFCPassword,
                              IN String Language,
                              IN Integer SecondLevelMonitor,
                              IN String MonitorProgram );

 Description : Validates the ArgList passed to the entry point.

 Parameters :

 $sEntryPointName               Entry pt name
 $localEnvFile                  Environments file for <LCUser>. 
 $localLCHome                   Directory for LiveCache database software.
 $localLCUser                   Operating system user for liveCache server.
 $localLCName                   Name of the liveCache instance.
 $localDBMUserId                User ID used to connect to liveCache server.
 $localDBMPassword              Password for the user <DBMUserId>.
 $localAPOHost                  Virtual hostname of the SAP APO Application Server.
 $localInstId                   SAP instance ID for the APO Application Server.
 $localClientId                 Production client ID for liveCache server.
 $localRFCDest                  RFC Destination for liveCache server start/stop.
 $localRFCHome                  Home directory for RFC SDK tool kit.
 $localRFCUserId                RFC User for SAP liveCache Server to operate globally.
 $localRFCPassword              Password for the <RFCUserId>.
 $localLanguage                 Language used to start/stop liveCache server.
 $localSecondLevelMonitor       Cycle count to run in-depth monitoring.
 $localMonitorProgram           Full path to external monitor program.
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub ArgsValid {
  #----------------------------------------
  # Set up local vars from passed in ones..
  #----------------------------------------
  my ( $sEntryPointName,
       $localEnvFile,
       $localLCHome,
       $localLCUser,
       $localLCName,
       $localDBMUserId,
       $localDBMPassword,
       $localAPOHost,
       $localInstId,
       $localClientId,
       $localRFCDest,
       $localRFCHome,
       $localRFCUserId,
       $localRFCPassword,
       $localLanguage,
       $localSecondLevelMonitor,
       $localMonitorProgram ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20001, "Subroutine <%s> called with:
  EntryPointName           [%s]
  EnvFile                  [%s]
  LCHome                   [%s]
  LCUser                   [%s]
  LCName                   [%s]
  DBMUserId                [%s]
  DBMPassword              [%s]
  APOHost                  [%s]
  InstId                   [%s]
  ClientId                 [%s]
  RFCDest                  [%s]
  RFCHome                  [%s]
  RFCUserId                [%s]
  RFCPassword              [%s]
  Language                 [%s]
  SecondLevelMonitor       [%s]
  MonitorProgram           [%s]",

  $sMyName,    $sEntryPointName,    $localEnvFile,    $localLCHome,
  $localLCUser,  $localLCName,  $localDBMUserId, $localDBMPassword,
  $localAPOHost,   $localInstId,   $localClientId,   $localRFCDest,
  $localRFCHome,         $localRFCUserId,        $localRFCPassword,
  $localLanguage, $localSecondLevelMonitor, $localMonitorProgram );

  #---------------------------------------------
  # Get the address of the Language parameter..
  #---------------------------------------------
  my $rsLanguage = \$_[14];

  #-------------------------------------
  # Set return code to hopeful success..
  #-------------------------------------
  my $iReturnCode = 1;

  #--------------------------------------------
  # Create hash to weed out empty args values..
  #--------------------------------------------
  my ( @lsMissing, %hsArgList ) = ();
  my @lsArgs =
  qw ( EntryPointName EnvFile LCHome LCUser LCName DBMUserId DBMPassword
       APOHost  InstId  ClientId  RFCDest  RFCHome RFCUserId RFCPassword
       Language SecondLevelMonitor MonitorProgram );

  @hsArgList { @lsArgs } = @_; # Assign..

  #----------------------------------------
  # Reqd. attributes based on entry point..
  #----------------------------------------
  my @lsOnlineArgs  = qw ( EnvFile LCHome LCUser LCName DBMUserId DBMPassword InstId );
  my @lsOfflineArgs = qw ( EnvFile LCHome LCUser LCName DBMUserId DBMPassword InstId );
  my @lsMonitorArgs = qw ( EnvFile LCHome LCUser LCName DBMUserId DBMPassword );
  my @lsCleanArgs   = qw ( EnvFile LCHome LCUser LCName DBMUserId DBMPassword );

  my %hlsNeeded = ( 'online'  => \@lsOnlineArgs,
                    'offline' => \@lsOfflineArgs,
                    'monitor' => \@lsMonitorArgs,
                    'clean'   => \@lsCleanArgs );

  my @lsReqdArgs = @{$hlsNeeded { $sEntryPointName }};

  #-----------------------
  # Check what's missing..
  #-----------------------
  if ( @lsMissing = grep { $hsArgList { $_ } !~ /\S/o } @lsReqdArgs ) {
    &logError ( 20002,
    "<%s> The following mandatory fields were not set [%s], Total [%s]",
    $sMyName, join ( ' ', @lsMissing ), scalar @lsMissing );
    $iReturnCode = 0; # Failure..
  }

  $iReturnCode || return $iReturnCode; # Return, if failed..

  #-----------------------------
  # Remove SLM file, if exists..
  #-----------------------------
  if ( $sEntryPointName ne 'monitor' ) {
    VCSagentFW::RunSecondLevelMonitor ( 1 ); # Delete..
  }

  #-------------------------------
  # Individual validation starts..
  #-------------------------------
  

  #-----------------------------------------------
  # Validation for online offline entry point(s)..
  #-----------------------------------------------
  if ( $sEntryPointName =~ /^(online|offline)$/o ) {
    #-----------------------------
    # Validate parameter EnvFile..
    #-----------------------------
    if ( ! -f  $localEnvFile ) {
      &logError ( 20003, "<%s> EnvFile [%s] is not a plain file",
      $sMyName, $localEnvFile );
      $iReturnCode = 0; # Failure..
    }

    #----------------------------
    # Validate parameter LCHome..
    #----------------------------
    if ( ! -d  $localLCHome ) {
      &logError ( 20005, "<%s> LCHome [%s] directory does not exist",
      $sMyName, $localLCHome );
      $iReturnCode = 0; # Failure..
    }

    #--------------------------
    # Validate parameter LCUser..
    #--------------------------
    if ( ! defined getpwnam $localLCUser ) {
      &logError ( 20007, "<%s> LCUser [%s] is not a valid UNIX login",
      $sMyName, $localLCUser );
      $iReturnCode = 0; # Failure..
    }

    #-----------------------------
    # Validate parameter InstId..
    #-----------------------------
    if ( $localInstId !~ /(^[0-8][0-9]$)/ &&  $localInstId !~ /(^9[0-6]$)/ ) {
      &logError ( 20008, "<%s> InstId [%s] not in the range of {00...96}.",
      $sMyName, $localInstId );
      $iReturnCode = 0; # Failure..
    }
   
    #-----------------------------
    # Validate parameter RFCHome..
    #-----------------------------
    if ( ( $localRFCHome =~ /(\S+)/ ) && ( ! -d  $localRFCHome ) ) {
      &logError ( 20004, "<%s> RFCHome [%s] directory does not exist",
      $sMyName, $localRFCHome );
      $iReturnCode = 0; # Failure..
    }
    #-----------------------------
    # Validate parameter APOHost..
    #-----------------------------
    if ( ( $localAPOHost  =~ /(\S+)/ ) && ( ! inet_aton $localAPOHost ) ) {
      &logError ( 20006, "<%s> APOHost [%s] is an invalid Hostname/IPV4 address",
      $sMyName, $localAPOHost );
      $iReturnCode = 0; # Failure..
    }
    #-----------------------------
    # Validate parameter ClientId..
    #-----------------------------
    if ( ( $localClientId  =~ /(\S+)/ ) && ( $localClientId !~ /(^[0-9][0-9][0-9]$)/ ) ) {
      &logError ( 20009, "<%s> ClientId [%s] not in the range of {000...999}.",
      $sMyName, $localClientId );
      $iReturnCode = 0; # Failure..
    }

    #-----------------------------
    # Validate parameter Language..
    #-----------------------------
    if ( ( $localLanguage  =~ /(\S+)/ ) && ( $localLanguage ne "EN" )) {
      &logWarn ( 20010, "<%s> Language [%s] is invalid. Defaulting the value to EN",
      $sMyName, $localLanguage );
      $$rsLanguage = "EN";
    }


  }

  #----------------------------------
  # Validation for all entry points..
  #----------------------------------

  #-----------------------------
  # Validate parameter LCName..
  #-----------------------------
  if ( $localLCName !~ /(^[A-Z][A-Z0-9][A-Z0-9]$)/ ) {
    &logError ( 20011, "<%s> LCName [%s] contains invalid characters",
    $sMyName, $localLCName );
    $iReturnCode = 0; # Failure..
  }


  #---------------------------------------
  # Validation for online entry point(s)..
  #---------------------------------------
  if ( $sEntryPointName eq "online" ) {
    #------------------------------------
    # Validate parameter MonitorProgram..
    #------------------------------------
    if ( $localMonitorProgram =~ /(\S+)/ && # Defined..
         ! -x $1 ) {
      &logError ( 20012, "<%s> MonitorProgram [%s] is not executable",
      $sMyName, $1 );
      $iReturnCode = 0; # Failure..
    }

  }

  $iReturnCode ||
  &logWarn ( 20013,
  "<%s> Some of the attributes could not be validated. Returning [%s]",
  $sMyName, $iReturnCode );

  return $iReturnCode ;
}



=pod

 -------------------------------------------------------------------------
 Function : bool Start ( IN String APOHost,
                         IN String ClientId,
                         IN String DBMPassword,
                         IN String DBMUserId,
                         IN String EnvFile,
                         IN String InstId,
                         IN String LCHome,
                         IN String LCName,
                         IN String LCUser,
                         IN String Language,
                         IN String RFCDest,
                         IN String RFCHome,
                         IN String RFCPassword,
                         IN String RFCUserId,
                         IN Integer SecondLevelMonitor );

 Description : Start logic needed for online entry pt

 Parameters :

 $sAPOHost                  Virtual hostname of the SAP APO Application Server.
 $sClientId                 Production client ID for liveCache server.
 $sDBMPassword              Password for the user <DBMUserId>.
 $sDBMUserId                User ID used to connect to liveCache server.
 $sEnvFile                  Environments file for <LCUser>. 
 $sInstId                   SAP instance ID for the APO Application Server.
 $sLCHome                   Directory for LiveCache database software.
 $sLCName                   Name of the liveCache instance.
 $sLCUser                   Operating system user for liveCache server.
 $sLanguage                 Language used to start/stop liveCache server.
 $sRFCDest                  RFC Destination for liveCache server start/stop.
 $sRFCHome                  Home directory for RFC SDK tool kit.
 $sRFCPassword              Password for the <RFCUserId>.
 $sRFCUserId                RFC User for SAP liveCache Server to operate globally.
 $iSecondLevelMonitor       Cycle count to run in-depth monitoring.
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub Start {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sAPOHost,
       $sClientId,
       $sDBMPassword,
       $sDBMUserId,
       $sEnvFile,
       $sInstId,
       $sLCHome,
       $sLCName,
       $sLCUser,
       $sLanguage,
       $sRFCDest,
       $sRFCHome,
       $sRFCPassword,
       $sRFCUserId,
       $iSecondLevelMonitor ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20014, "Subroutine <%s> called with:
  APOHost                  [%s]
  ClientId                 [%s]
  DBMPassword              [%s]
  DBMUserId                [%s]
  EnvFile                  [%s]
  InstId                   [%s]
  LCHome                   [%s]
  LCName                   [%s]
  LCUser                   [%s]
  Language                 [%s]
  RFCDest                  [%s]
  RFCHome                  [%s]
  RFCPassword              [%s]
  RFCUserId                [%s]
  SecondLevelMonitor       [%s]",

  $sMyName,  $sAPOHost,  $sClientId,  $sDBMPassword,   $sDBMUserId,
  $sEnvFile,  $sInstId,  $sLCHome,  $sLCName, $sLCUser, $sLanguage,
  $sRFCDest,      $sRFCHome,      $sRFCPassword,       $sRFCUserId,
  $iSecondLevelMonitor );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1;

  #-----------------------------------------
  # Whack any preexisting SAPLiveCache processes..
  #-----------------------------------------
  &logTrace ( 20015,
  "<%s> Killing SAPLiveCache processes for this instance", $sMyName );

  if ( SAPLiveCache::KillSAPLiveCacheProcesses ( $sLCName,
                                                 $sLCUser ) ) {
    &logTrace ( 20016,
    "<%s> Processes for this instance do not exist before start-up",
    $sMyName );
  }
  else {
    &logError ( 20017,
    "<%s> Could not remove all processes for this instance. Exiting",
    $sMyName );
    $iReturnCode = 0; # Failure..
  }

  #--------------------
  # Return, if failed..
  #--------------------
  $iReturnCode || return $iReturnCode;

  #------------------------------------------------------
  # Remove any preexisting SAPLiveCache IPC resources...
  #------------------------------------------------------
  &logTrace ( 20018,
  "<%s> Cleaning SAPLiveCache IPC resources for instance [%s]", $sMyName, $sLCName );

  if(SAPLiveCache::RemoveIPCResourcesForSAPLiveCache ( "sdb",
                                                       $sLCHome,
                                                       $sLCName ) ) {
    &logTrace(20019, 
              "<%s> Removed all IPC resources for SAP liveCache Instance [%s] successfully",
              $sMyName, $sLCName );
  }
  else {
    &logError(20020, 
              "<%s> Failed to remove IPC resources for SAP liveCache Instance [%s]. Remove all IPC resources for this instance manually before retrying online", $sMyName, $sLCName );
    $iReturnCode = 0; # Failure ..
  }
  
  #--------------------
  # Return, if failed..
  #--------------------
  $iReturnCode || return $iReturnCode;

  #---------------------------------------------------------------------
  # Upper bound.. Time in EPOCH when entry pt started + VCS time limit..
  #---------------------------------------------------------------------
  my $iTillWhen = $^T + SAPLiveCache::GetEngineTimeout ( 1 );

  my ( @lsCmdout, $iExitCode, $iTimeout, @liPids ) = ();
  
  #-------------------------------------------
  # Start x_server for SAP liveCache Server..
  #-------------------------------------------
  my $sCommand = "$sLCHome/x_server";
  my $sCommandArgs = "start";

  &logTrace ( 20021,
  "<%s> Starting x_server with Cmd [%s] Args [%s]",
  $sMyName, $sCommand, $sCommandArgs );

  #------------------
  # Request startup..
  #------------------
  if ( ( $iTimeout = $iTillWhen - time ) <= 10 ) { # Get time bearings..
    &logError ( 20022,
    "<%s> Insufficient time to start x_server. Short by [%s] sec",
    $sMyName, abs ( $iTimeout ) );
    $iReturnCode = 0; # Failure..
  }
  elsif ( $iReturnCode =
    SAPLiveCache::ExecLocalCmd ( $sLCUser,
                                 $sEnvFile,  
                                 $sCommand,
                                 $sCommandArgs,
                                 \$iExitCode,
                                 $iTimeout,
                                 \@lsCmdout ) ) {
    &logTrace ( 20023,
    "<%s> Cmd [%s] with args [%s] executed successfully",
    $sMyName, $sCommand, $sCommandArgs );

    my $sSrchStr = qr|\bXSERVER\b.*\bstarted\b|io; # Search pattern..
    my $sSrchStr1 = qr|\bFound\s+running\s+XServer\b|o; # Search pattern..

    &logTrace ( 20024, "<%s> Looking with filter [%s] or [%s] from output",
    $sMyName, $sSrchStr, $sSrchStr1 );

    #--------------------------------------------
    # Attempt to locate search string in output..
    #--------------------------------------------
    unless ( grep { /$sSrchStr/ || /$sSrchStr1/ } @lsCmdout ) {
      &logError ( 20025,
      "<%s> Failed to locate search pattern [%s] or [%s]  in Cmd output [%s]",
      $sMyName, $sSrchStr, $sSrchStr1, join ( chr 012, @lsCmdout ) );

      $iReturnCode = 0; # Failure..
    }
  }
  else {
    &logError ( 20026,
    "<%s> Cmd [%s] with args [%s] failed with exit code [%s].
    Cmd output [%s]",
    $sMyName, $sCommand, $sCommandArgs, $iExitCode,
    join ( chr 012, @lsCmdout ) );
    $iReturnCode = 0; # Failure..
  }

  #-------------------
  # Return if failed..
  #-------------------
  $iReturnCode || return $iReturnCode;

  #---------------------------------------------------------------------
  # Upper bound.. Time in EPOCH when entry pt started + VCS time limit..
  #---------------------------------------------------------------------
  #$iTillWhen = $^T + SAPLiveCache::GetEngineTimeout ( 1 );
  
  #------------------------------------------
  # Declare version command and args variables..
  #------------------------------------------
  my $lcVersion = 0;

  $sCommand = "$sLCHome/dbmcli";
  $sCommandArgs = "-s inst_enum";

  &logInfo ( 20234,
  "<%s> Attempting to get the version of SAPLiveCache with Command [%s] Args [%s]",
  $sMyName, $sCommand, $sCommandArgs );

  #------------------------
  # Request version number.
  #------------------------

  #-------------------------------------------
  # Declare start command and args variables..
  #-------------------------------------------
  if ( ( $iTimeout = $iTillWhen - time ) <= 10 ) { # Get time bearings..
    &logError ( 20235,
    "<%s> Insufficient time to get version number of livecache database. Short by [%s] sec",
    $sMyName, abs ( $iTimeout ) );

    $iReturnCode = 0; # Failure..
  }
  elsif ( SAPLiveCache::ExecLocalCmd ( $sLCUser,
                                       $sEnvFile,
                                       $sCommand,
                                       $sCommandArgs,
                                       \$iExitCode,
                                       $iTimeout,
                                       \@lsCmdout,
                                       $sDBMPassword ) ) {
    &logTrace ( 20236,
    "<%s> Cmd [%s] with args [%s] executed successfully",
    $sMyName, $sCommand, $sCommandArgs );
  }
  
  #------------------------------------------------------
  # Get version of livecache database.
  #------------------------------------------------------
  foreach (  @lsCmdout ) {
    if( $_ =~ /\b(\S*)\s*(\S+$sLCName)\b/io ) {
      my ( $version, $path ) = ($1, $2);
      $version =~ s/(\d+\.\d+).*/$1/;
      $lcVersion = $version ;
    }
  }

  @lsCmdout = (); # Free..
	
	
  my $sPassword = () ;
  my $sSAPSid = () ;
  my $sSapEvtCmd = () ;
  if( $sLCUser  =~ /\b(\S*)adm\b/io ) {
    $sSAPSid = uc $1 ;
    $sSapEvtCmd = "/usr/sap/$sSAPSid/SYS/exe/run/sapevt";
  }	
  if ( ( ( -e $sSapEvtCmd ) && (-x $sSapEvtCmd ) ) && 
       ( ( $sRFCHome !~ /(\S+)/ ) || ( $sAPOHost  !~ /(\S+)/ ) || ( $sClientId  !~ /(\S+)/ ) || ( $sRFCUserId  !~ /(\S+)/ ))  ) {
    $sCommand = $sSapEvtCmd ;
    $sCommandArgs = " Z_PLEASE_START_LIVECACHE name=$sSAPSid nr=$sInstId";
    $sPassword = undef ;
  }
  elsif ($lcVersion >= 7.9) { 
    $sCommand = "$sLCHome/dbmcli";
    $sCommandArgs = "-d $sLCName -u $sDBMUserId,$sDBMPassword db_online";
    $sPassword = $sDBMPassword ;
  }
  else {
    $sCommand = "$sRFCHome/startrfc";
    $sCommandArgs = "-3 -d $sRFCDest -h $sAPOHost -s $sInstId " .
                  "-c $sClientId -u $sRFCUserId -p $sRFCPassword " .
                  "-l $sLanguage -F START_LIVECACHE_LVC -E IV_CON_NAME=LCA";
    $sPassword = $sRFCPassword ;
  }
  

  ( @lsCmdout, $iExitCode, $iTimeout, @liPids ) = ();


  &logTrace ( 20027,
  "<%s> Starting SAPLiveCache instance with Cmd [%s] Args [%s]",
  $sMyName, $sCommand, $sCommandArgs );

  #------------------
  # Request startup..
  #------------------
  if ( ( $iTimeout = $iTillWhen - time ) <= 10 ) { # Get time bearings..
    &logError ( 20028,
    "<%s> Insufficient time to continue boot up. Short by [%s] sec",
    $sMyName, abs ( $iTimeout ) );
    $iReturnCode = 0; # Failure..
  }
  elsif ( $iReturnCode =
    SAPLiveCache::ExecLocalCmd ( $sLCUser, 
                                 $sEnvFile, 
                                 $sCommand, 
                                 $sCommandArgs, 
                                 \$iExitCode, 
                                 $iTimeout, 
                                 \@lsCmdout, 
                                 $sPassword ) ) {
    &logTrace ( 20029,
    "<%s> Cmd [%s] with args [%s] executed successfully",
    $sMyName, $sCommand, $sCommandArgs );
  }
  else {
    &logError ( 20030,
    "<%s> Cmd [%s] with args [%s] failed with exit code [%s]. 
    Cmd output [%s]",
    $sMyName, $sCommand, $sCommandArgs, $iExitCode,
    join ( chr 012, @lsCmdout ) );
    $iReturnCode = 0; # Failure..
  }

  #-------------------
  # Return if failed..
  #-------------------
  $iReturnCode || return $iReturnCode;

  #--------------------
  # Delay entry point..
  #--------------------
  my $iSleepTime = 10; # Secs..
  
  while ( ( $iTimeout = $iTillWhen - time ) > 10 ) {
    @liPids = (); # Initialise..

    sleep $iSleepTime; # Sec..

    &logTrace ( 20031,
    "<%s> Confirming start-up with [%s] sec in hand",
    $sMyName, $iTimeout );

    #---------
    # Do FLM..
    #---------
    if ( VCSDef_MonitorExit::ONLINE ==
         SAPLiveCache::GetProcState ( \@liPids,
                                      $sLCName,
                                      $sLCUser,
                                      1 ) ) {      # Running FFLM..
      #---------------------
      # Start-up confirmed..
      #---------------------
      $iReturnCode = 1; # Success..
      last;
    }
    else {
      &logTrace ( 20032,
      "<%s> Could not confirm start-up. Sleeping for [%s] sec",
      $sMyName, $iSleepTime );

      $iReturnCode = 0;  # Failure..
    }
  }

  if ( $iReturnCode && $iSecondLevelMonitor ) {  # If enabled..
    while ( ( $iTimeout = $iTillWhen - time ) > 10 ) {
      sleep $iSleepTime; # Sec..

      &logTrace ( 20033,
      "<%s> Confirming start-up with [%s] sec in hand",
      $sMyName, $iTimeout );

      #---------
      # Do SLM..
      #---------
      if ( VCSDef_MonitorExit::ONLINE ==                       # Online after..
           SAPLiveCache::SecondLevelMonitor ( $sDBMPassword,   # Running SLM..
                                              $sDBMUserId,
                                              $sEnvFile,
                                              $sLCHome,
                                              $sLCName,
                                              $sLCUser,
                                              $iTillWhen ) ) { 
        #---------------------
        # Start-up confirmed..
        #---------------------
        &logInfo ( 20034,
        "<%s> Confirmed start-up of this instance of SAPLiveCache server",
        $sMyName );

        $iReturnCode = 1; # Success..
        last;
      }
      else {
        &logTrace ( 20035,
        "<%s> Could not confirm start-up. Sleeping for [%s] sec",
        $sMyName, $iSleepTime );
      }
    }
  }
  elsif( $iReturnCode && !$iSecondLevelMonitor ) {   # If only FLM succeeds ..
    &logInfo ( 20036,
    "<%s> Confirmed start-up of this instance of SAPLiveCache server",
    $sMyName );
  }


  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool Stop ( IN String APOHost,
                        IN String ClientId,
                        IN String EnvFile,
                        IN String InstId,
                        IN String LCHome,
                        IN String LCName,
                        IN String LCUser,
                        IN String Language,
                        IN String RFCDest,
                        IN String RFCHome,
                        IN String RFCPassword,
                        IN String RFCUserId 
                        IN String DBMPassword,
                        IN String DBMUserId);

 Description : Stop logic needed for offline entry pt

 Parameters :

 $sAPOHost                  Virtual hostname of the SAP APO Application Server.
 $sClientId                 Production client ID for liveCache server.
 $sEnvFile                  Environments file for <LCUser>. 
 $sInstId                   SAP instance ID for the APO Application Server.
 $sLCHome                   Directory for LiveCache database software.
 $sLCName                   Name of the liveCache instance.
 $sLCUser                   Operating system user for liveCache server.
 $sLanguage                 Language used to start/stop liveCache server.
 $sRFCDest                  RFC Destination for liveCache server start/stop.
 $sRFCHome                  Home directory for RFC SDK tool kit.
 $sRFCPassword              Password for the <RFCUserId>.
 $sRFCUserId                RFC User for SAP liveCache Server to operate globally.
 $sDBMPassword              Password for the user <DBMUserId>.
 $sDBMUserId                User ID used to connect to liveCache server.
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub Stop {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sAPOHost,
       $sClientId,
       $sEnvFile,
       $sInstId,
       $sLCHome,
       $sLCName,
       $sLCUser,
       $sLanguage,
       $sRFCDest,
       $sRFCHome,
       $sRFCPassword,
       $sRFCUserId,
       $sDBMPassword,
       $sDBMUserId ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20037, "Subroutine <%s> called with:
  APOHost                  [%s]
  ClientId                 [%s]
  EnvFile                  [%s]
  InstId                   [%s]
  LCHome                   [%s]
  LCName                   [%s]
  LCUser                   [%s]
  Language                 [%s]
  RFCDest                  [%s]
  RFCHome                  [%s]
  RFCPassword              [%s]
  RFCUserId                [%s]
  DBMPassword              [%s]
  DBMUserId                [%s]",

  $sMyName, $sAPOHost, $sClientId, $sEnvFile,  $sInstId,  $sLCHome, $sLCName,
  $sLCUser,   $sLanguage,   $sRFCDest,   $sRFCHome,  $sRFCPassword, $sRFCUserId,$sDBMPassword,   $sDBMUserId );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1; # Success..

  #---------------------------------------------------------------------
  # Upper bound.. Time in EPOCH when entry pt started + VCS time limit..
  #---------------------------------------------------------------------
  my $iTillWhen = $^T + SAPLiveCache::GetEngineTimeout ( 1 ); # With margin..

  #------------------------------------------
  # Declare stop command and args variables..
  #------------------------------------------
  my ( $sCommand, $sCommandArgs, $iTimeout, $iExitCode, $iSleepTime,
       @lsCmdout, @liPids ) = ();
  
  #------------------------------------------
  # Declare version command and args variables..
  #------------------------------------------
  my $lcVersion = 0;

  $sCommand = "$sLCHome/dbmcli";
  $sCommandArgs = "-s inst_enum";

  &logInfo ( 20237,
  "<%s> Attempting to get the version of  SAPLiveCache with Command [%s] Args [%s]",
  $sMyName, $sCommand, $sCommandArgs );

  #------------------------
  # Request version number.
  #------------------------

  #-------------------------------------------
  # Declare start command and args variables..
  #-------------------------------------------
  if ( ( $iTimeout = $iTillWhen - time ) <= 10 ) { # Get time bearings..
    &logError ( 20238,
    "<%s> Insufficient time to get version number of livecache database. Short by [%s] sec",
    $sMyName, abs ( $iTimeout ) );

    $iReturnCode = 0; # Failure..
  }
  elsif ( SAPLiveCache::ExecLocalCmd ( $sLCUser,
                                       $sEnvFile,
                                       $sCommand,
                                       $sCommandArgs,
                                       \$iExitCode,
                                       $iTimeout,
                                       \@lsCmdout,
                                       $sDBMPassword ) ) {
    &logTrace ( 20239,
    "<%s> Cmd [%s] with args [%s] executed successfully",
    $sMyName, $sCommand, $sCommandArgs );
  }
  
  #------------------------------------------------------
  # Get version of livecache database.
  #------------------------------------------------------
  foreach (  @lsCmdout ) {
    if( $_ =~ /\b(\S*)\s*(\S+$sLCName)\b/io ) {
      my ( $version, $path ) = ($1, $2);
      $version =~ s/(\d+\.\d+).*/$1/;
      $lcVersion = $version ;
    }
  }

  @lsCmdout = (); # Free..
	
	
  my $sPassword = () ;
  my $sSAPSid = () ;
  my $sSapEvtCmd = () ;
  if( $sLCUser  =~ /\b(\S*)adm\b/io ) {
    $sSAPSid = uc $1 ;
    $sSapEvtCmd = "/usr/sap/$sSAPSid/SYS/exe/run/sapevt";
  }
  if ( ( ( -e $sSapEvtCmd ) && (-x $sSapEvtCmd ) ) && 
       ( ( $sRFCHome !~ /(\S+)/ ) || ( $sAPOHost  !~ /(\S+)/ ) || ( $sClientId  !~ /(\S+)/ ) || ( $sRFCUserId  !~ /(\S+)/ ))  ) {
    $sCommand = $sSapEvtCmd ;
    $sCommandArgs = " Z_PLEASE_STOP_LIVECACHE name=$sSAPSid nr=$sInstId";
    $sPassword = undef ;
  }
  elsif ($lcVersion >= 7.9) { 
    $sCommand = "$sLCHome/dbmcli";
    $sCommandArgs = "-d $sLCName -u $sDBMUserId,$sDBMPassword db_offline";
    $sPassword = $sDBMPassword ;
  }
  else {
  $sCommand = "$sRFCHome/startrfc";
  $sCommandArgs = "-3 -d $sRFCDest -h $sAPOHost -s $sInstId " .
                  "-c $sClientId -u $sRFCUserId -p $sRFCPassword " .
                  "-l $sLanguage -F STOP_LIVECACHE_LVC -E IV_CON_NAME=LCA";
  $sPassword = $sRFCPassword ;
  }

  &logTrace ( 20038,
  "<%s> Stopping SAPLiveCache with Command [%s] Args [%s]",
  $sMyName, $sCommand, $sCommandArgs );

  #------------------------
  # Request shutdown..
  #------------------------
  if ( ( $iTimeout = $iTillWhen - time ) <= 10 ) { # Get time bearings..
    &logError ( 20039,
    "<%s> Insufficient time to continue shutdown. Short by [%s] sec",
    $sMyName, abs ( $iTimeout ) );
    $iReturnCode = 0; # Failure..
  }
  elsif ( $iReturnCode =
          SAPLiveCache::ExecLocalCmd ( $sLCUser, 
                                       $sEnvFile, 
                                       $sCommand, 
                                       $sCommandArgs, 
                                       \$iExitCode, 
                                       $iTimeout, 
                                       \@lsCmdout, 
                                       $sPassword ) ) {
    &logTrace ( 20040,
    "<%s> Cmd [%s] with args [%s] executed successfully",
    $sMyName, $sCommand, $sCommandArgs );

    #-----------------
    # Set sleep time..
    #-----------------
    $iSleepTime = 10; # Sec..
  
    &logInfo ( 20041,
    "<%s> Delaying exit to confirm shutdown..", $sMyName );

    #--------------------
    # Delay entry point..
    #--------------------
    while ( ( $iTimeout = $iTillWhen - time ) > 20 ) {
      #-----------------------
      # Wait before retrying..
      #-----------------------
      sleep $iSleepTime; # Secs..
  
      SAPLiveCache::GetProcState ( \@liPids,
                                   $sLCName,
                                   $sLCUser,
                                   0 ); # Request regular..
  
      #------------------------------
      # Check for pending processes..
      #------------------------------
      if ( scalar @liPids ) {
        &logTrace ( 20042,
        "<%s> Could not confirm shutdown. Retrying in [%s] sec.
        Delaying kill with [%s] sec in hand",
        $sMyName, $iSleepTime, $iTimeout );
      }
      else {
        &logInfo ( 20043,
        "<%s> Confirmed shutdown of SAPLiveCache instance",
        $sMyName );
        last; # Get out..
      }
    }
  }
  else {
    &logError ( 20044,
    "<%s> Cmd [%s] with args [%s] failed to stop the liveCache Server.",
    $sMyName, $sCommand, $sCommandArgs );

    $iReturnCode = 0; # Failure..
  }


  #------------------------------------------------
  # Whack any preexisting SAPLiveCache processes..
  #------------------------------------------------
  &logTrace ( 20045,
  "<%s> Killing SAPLiveCache processes for this instance", $sMyName );

  if ( SAPLiveCache::KillSAPLiveCacheProcesses ( $sLCName,
                                                 $sLCUser ) ) {
    &logTrace ( 20046,
    "<%s> Processes for this instance do not exist after shutdown",
    $sMyName );
    $iReturnCode = 1; # Success ..
  }
  else {
    &logError ( 20047,
    "<%s> Could not remove all processes for this instance. Exiting",
    $sMyName );
    $iReturnCode = 0; # Failure..
  }
 
  #------------------------------------------------------
  # Remove any left over SAPLiveCache IPC resources...
  #------------------------------------------------------
  &logTrace ( 20048,
  "<%s> Cleaning SAPLiveCache IPC resources for instance [%s]", $sMyName, $sLCName );

  if(SAPLiveCache::RemoveIPCResourcesForSAPLiveCache ( "sdb",
                                                       $sLCHome,
                                                       $sLCName ) ) {
    &logTrace(20049, 
              "<%s> Removed all IPC resources for SAP liveCache Instance [%s] successfully",
              $sMyName, $sLCName );
  }
  else {
    &logError(20050, 
              "<%s> Removed all IPC resources for SAP liveCache Instance [%s] successfully",
              $sMyName, $sLCName );
    $iReturnCode = 0; # Failure ..
  }
  
  return $iReturnCode;
}



=pod

 -------------------------------------------------------------------------
 Function : bool Kill ( IN Integer CleanReason,
                        IN String DBMPassword,
                        IN String DBMUserId,
                        IN String EnvFile,
                        IN String LCHome,
                        IN String LCName,
                        IN String LCUser );

 Description : Forces shutdown of resource needed for clean entry point

 Parameters :

 $iCleanReason              Why clean was called
 $sDBMPassword              Password for the user <DBMUserId>.
 $sDBMUserId                User ID used to connect to liveCache server.
 $sEnvFile                  Environments file for <LCUser>. 
 $sLCHome                   Directory for LiveCache database software.
 $sLCName                   Name of the liveCache instance.
 $sLCUser                   Operating system user for liveCache server.
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub Kill {
  #----------------
  # Get Arguments..
  #----------------
  my ( $iCleanReason,
       $sDBMPassword,
       $sDBMUserId,
       $sEnvFile,
       $sLCHome,
       $sLCName,
       $sLCUser ) = @_;

  #-----------------------------
  # Figure out why were called..
  #-----------------------------
  my ( $sCleanReason, %hsCleanReasons ) = ();
  my @lsCleanStrings = qw ( Offline_timed_out    Offline_was_ineffective
                            Online_timed_out      Online_was_ineffective
                            Found_offline_unexpectedly Monitor_timed_out );

  @hsCleanReasons { 0 .. $#lsCleanStrings } = @lsCleanStrings; # Populate..
  ( $sCleanReason = $hsCleanReasons { $iCleanReason } ||       # Reason..
                    $iCleanReason ) =~ s/_/ /g;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20051, "Subroutine <%s> called with:
  CleanReason              [%s]
  DBMPassword              [%s]
  DBMUserId                [%s]
  EnvFile                  [%s]
  LCHome                   [%s]
  LCName                   [%s]
  LCUser                   [%s]",

  $sMyName, $sCleanReason, $sDBMPassword,  $sDBMUserId,  $sEnvFile,
  $sLCHome, $sLCName, $sLCUser );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1;

  #---------------------------------------------------------------------
  # Upper bound.. Time in EPOCH when entry pt started + VCS time limit..
  #---------------------------------------------------------------------
  my $iTillWhen = $^T + SAPLiveCache::GetEngineTimeout ( 1 ); # With margin..

  #------------------------------------------
  # Declare stop command and args variables..
  #------------------------------------------
  my ( $sCommand, $sCommandArgs, $iTimeout, $iExitCode,
       @lsFiles, @lsCmdout ) = ();
  
  $sCommand = "$sLCHome/dbmcli";
  $sCommandArgs = "-d $sLCName -u $sDBMUserId,$sDBMPassword db_offline";

  unless ( $hsCleanReasons { $iCleanReason } eq 'Offline_timed_out' ||
           $hsCleanReasons { $iCleanReason } eq 'Offline_was_ineffective' ) {
    #-------------------
    # Try to be gentle..
    #-------------------
    &logInfo ( 20052,
    "<%s> Attempting to stop SAPLiveCache with Command [%s] Args [%s]",
    $sMyName, $sCommand, $sCommandArgs );
  
    #------------------------
    # Request shutdown..
    #------------------------
    if ( ( $iTimeout = $iTillWhen - time ) <= 10 ) { # Get time bearings..
      &logError ( 20053,
      "<%s> Insufficient time to continue shutdown. Short by [%s] sec",
      $sMyName, abs ( $iTimeout ) );

      $iReturnCode = 0; # Failure..
    }
    elsif ( SAPLiveCache::ExecLocalCmd ( $sLCUser, 
                                         $sEnvFile, 
                                         $sCommand, 
                                         $sCommandArgs, 
                                         \$iExitCode, 
                                         $iTimeout, 
                                         \@lsCmdout, 
                                         $sDBMPassword ) ) {
      &logTrace ( 20054, 
      "<%s> Cmd [%s] with args [%s] executed successfully",
      $sMyName, $sCommand, $sCommandArgs );
    }
  }  # End of unless block.

  #------------------------------------------------
  # Whack any preexisting SAPLiveCache processes..
  #------------------------------------------------
  &logTrace ( 20055,
  "<%s> Killing SAPLiveCache processes for this instance", $sMyName );

  if ( SAPLiveCache::KillSAPLiveCacheProcesses ( $sLCName,
                                                 $sLCUser ) ) {
    &logTrace ( 20056,
    "<%s> Processes for this instance do not exist after shutdown",
    $sMyName );
  }
  else {
    &logError ( 20057,
    "<%s> Could not remove all processes for this instance. Exiting",
    $sMyName );
    $iReturnCode = 0; # Failure..
  }

  #-------------------------------------------
  # Define clear command and args variables..
  #-------------------------------------------
  $sCommand = "$sLCHome/dbmcli";
  $sCommandArgs = "-d $sLCName -u $sDBMUserId,$sDBMPassword db_clear";

  #-------------------
  # Try to be gentle..
  #-------------------
  &logInfo ( 20058,
  "<%s> Attempting to clear SAPLiveCache with Command [%s] Args [%s]",
  $sMyName, $sCommand, $sCommandArgs );

  #-----------------
  # Request clear ..
  #-----------------
  if ( ( $iTimeout = $iTillWhen - time ) <= 10 ) { # Get time bearings..
    &logError ( 20059,
    "<%s> Insufficient time to continue clear. Short by [%s] sec",
    $sMyName, abs ( $iTimeout ) );
    #$iReturnCode = 0; # Failure..
  }
  elsif ( SAPLiveCache::ExecLocalCmd ( $sLCUser,
                                       $sEnvFile,
                                       $sCommand,
                                       $sCommandArgs,
                                       \$iExitCode,
                                       $iTimeout,
                                       \@lsCmdout,
                                       $sDBMPassword ) ) {
    &logTrace ( 20060,
    "<%s> Cmd [%s] with args [%s] executed successfully",
    $sMyName, $sCommand, $sCommandArgs );
  }

  #------------------------------------------------------
  # Remove any left over SAPLiveCache IPC resources...
  #------------------------------------------------------
  &logTrace ( 20061,
  "<%s> Cleaning SAPLiveCache IPC resources for instance [%s]", $sMyName, $sLCName );

  if( $iReturnCode = SAPLiveCache::RemoveIPCResourcesForSAPLiveCache ( "sdb",
                                                                        $sLCHome,
                                                                        $sLCName ) ) {
    &logTrace(20062, 
              "<%s> Removed all IPC resources for SAP liveCache Instance [%s] successfully",
              $sMyName, $sLCName );
  }
  else {
    &logError(20063, 
              "<%s> Failed to Remove all IPC resources for SAP liveCache Instance [%s]",
              $sMyName, $sLCName );
    $iReturnCode = 0; # Failure ..
  }

  return $iReturnCode;
}



=pod

 -------------------------------------------------------------------------
 Function : int GetResourceState ( IN String LCName,
                                   IN String LCUser );

 Description : First level monitor to determine state of resource.

 Parameters :

 $sLCName                   Name of the liveCache instance.
 $sLCUser                   Operating system user for liveCache server.
 
 Returns : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 105  ONLINE with confidence level 50   VCSDef_MonitorExit::PARTIAL_ONLINE
 110  ONLINE with confidence level 100          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 -------------------------------------------------------------------------

=cut

sub GetResourceState {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sLCName,
       $sLCUser ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20064, "Subroutine <%s> called:
  LCName                   [%s]
  LCUser                   [%s]",

  $sMyName, $sLCName, $sLCUser );

  my @liDummyPids = ();
  my $iReturnCode = SAPLiveCache::GetProcState ( \@liDummyPids,
                                                 $sLCName,
                                                 $sLCUser,
                                                 1 ); # Use FFLM..

  if ( VCSDef_MonitorExit::ONLINE == $iReturnCode ) {
    &logTrace ( 20065,
    "<%s> Found all required processes. Return code set to [%s]",
    $sMyName, $iReturnCode );
  }
  else {
    &logTrace ( 20066,
    "<%s> Did not find all required processes. Returning [%s]",
    $sMyName, $iReturnCode );
  }

  #--------------------------------
  # Populate the global PID array..
  #--------------------------------
  @liRunningProcs = @liDummyPids;
                                                                                
  &logTrace ( 20067,
  "<%s> Populating global running proclist with [%s] PIDs [%s]",
  $sMyName, scalar @liRunningProcs, join ( ' ', @liRunningProcs ) );

  return ( $iReturnCode );
}


=pod

 -------------------------------------------------------------------------
 Function : int GetProcState ( OUT ListReference OnlinePids,
                               IN String LCName,
                               IN String LCUser,
                               IN Integer UseFFLM );

 Description : First level monitor to determine state of resource.

 Parameters :

 $rliOnlinePids             PIDs of relevant running processes
 $sLCName                   Name of the liveCache instance.
 $sLCUser                   Operating system user for liveCache server.
 $iUseFFLM                  Filtering procedure
 
 Returns : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 105  ONLINE with confidence level 50   VCSDef_MonitorExit::PARTIAL_ONLINE
 110  ONLINE with confidence level 100          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 -------------------------------------------------------------------------

=cut

sub GetProcState {
  #----------------
  # Get Arguments..
  #----------------
  my ( $rliOnlinePids,
       $sLCName,
       $sLCUser,
       $iUseFFLM ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20068, "Subroutine <%s> called with:
  OnlinePids               [%s]
  LCName                   [%s]
  LCUser                   [%s]
  UseFFLM                  [%s]",

  $sMyName, $rliOnlinePids, $sLCName, $sLCUser, $iUseFFLM );

  #-----------------------------
  # Set return code to success..
  #-----------------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE; # Success..

  $sLCUser = "sdb"; # Change the user to liveCache process owner..

  #----------------------
  # Initialise Variables..
  #----------------------
  my ( $iProcessCount, $iUseRFLM, $iFound, $iUid, $iPid, $sLang,
       $sKey, $rhlProc, $sPidFile, $sProc, $sTemp,
       @hProcListHash, @lsUpdates, @lsValues,
       %hiProcPids, %hsSrvrDetails, %hsCopy ) = ();

  @$rliOnlinePids = ();
  my %hiLCPid = ();
  my $iPPid = ();

  #------------------------------------
  # Attempt to get monitoring details..
  #------------------------------------
  if ( ( 0 == ( $iUid = getpwnam $sLCUser ) ) && # User is the super-user..
       ( 2 == $iUseFFLM ) ) {                  # And called to kill-by-UID..
    &logTrace ( 20069, "<%s> Unsafe to kill processes by User [%s] UID [%s].
    Using filter-based approach.", $sMyName, $sLCUser, $iUid ),
    $iUseFFLM = 1;
  }

  if ( $iUseFFLM != 2 ) { # Called for first level check..
    if ( SAPLiveCache::GetSrvrDetails ( $sLCName, 
                                        \%hsSrvrDetails ) ) {
      &logTrace ( 20070,
      "<%s> Got details for monitoring SAPLiveCache instance", $sMyName );
  
      %hsCopy = %hsSrvrDetails;        # Backup..
      $iUseRFLM = scalar keys %hsCopy; # Set to entire count..
    }
    else {
      &logError ( 20071,
      "<%s> Failed to get details for monitoring SAPLiveCache instance",
      $sMyName );
      return VCSDef_MonitorExit::UNKNOWN;
    }
                                  
    #---------------------------------
    # Check if we attempt to be fast.. 
    #---------------------------------
    if ( $iUseFFLM ) { # FFLM..
      $iUseRFLM = 0;   # Hopefully skip..

      foreach $sKey ( keys %hsCopy ) {          # Use copy..
        @lsUpdates = ();                        # Reset..
        @lsValues = @{ $hsCopy { $sKey } };     # Get values..
        $sTemp .= "\nSAPLiveCache process [$sKey]\n"; # Display string..

        #----------------
        # Read in pairs..
        #----------------
        while ( ( $sPidFile, $sProc ) = splice ( @lsValues, 0, 2 ) ) {

          #-----------
          # Run FFLM..
          #-----------
          $iFound = SAPLiveCache::FFLMStatus ( $sPidFile,
                                               $sProc,
                                               $iUid,
                                               \$iPid, 
                                               \$iPPid ); # by vchappav ...
          #----------------
          # Update status..
          #----------------
          push ( @lsUpdates, $sPidFile, $sProc, $iFound );

          $sTemp .= "PID File     [$sPidFile]\n";
          $sTemp .= "Srch Process [$sProc]\n";
          $sTemp .= "FFLM Status  [$iFound]\n\n";
        }

        @{ $hsCopy { $sKey } } = @lsUpdates;

        #------------------------
        # Handle monitor status..
        #------------------------
        if ( VCSDef_MonitorExit::ONLINE == $iFound ) {
          push ( @$rliOnlinePids, $iPid ); # Push..
          $hiProcPids { $sKey } = $iPid;   # Store..

          $hiLCPid { $iPid } = $iPPid;    # by vchappav ...
        }
        else {
          ++ $iUseRFLM; # Mark it for RFLM..
        }
      }

      #----------------
      # Print summary..
      #----------------
      &logTrace ( 20072,
      "<%s> SAPLiveCache instance process details ---%s\n\n",
      $sMyName, $sTemp );

      ( $sTemp, $sPidFile, $sProc ) = (); # Release memory..
    }
    else { # Use RLFM..
      #-----------------------------------------
      # Mark status of each service as offline..
      #-----------------------------------------
      $iFound = VCSDef_MonitorExit::OFFLINE; # Set to offline..
      foreach $sKey ( keys %hsCopy ) {
        @lsUpdates = (); # Reset..

        @lsValues = @{ $hsCopy { $sKey } }; # Get values..
        while ( scalar @lsValues ) {
          push ( @lsUpdates, splice ( @lsValues, 0, 2 ), $iFound );
        }

        @{ $hsCopy { $sKey } } = @lsUpdates; # Update..
      }
    }
  }
  else { # Called for Kill..
    &logTrace ( 20073,
    "<%s> Called to get processes for User [%s] UID [%s]",
    $sMyName, $sLCUser, $iUid );
    ++ $iUseRFLM;
  }

  #-----------------------
  # Regular slow :-( FLM..
  #-----------------------
  if ( $iUseRFLM ) {
    &logTrace ( 20074,
    "<%s> Running regular first level monitor", $sMyName );

    #----------------------------------
    # Set LANG for better performance..
    #----------------------------------
    $sLang = $ENV { LANG }; # Backup..
    $ENV { LANG } = 'C';    # Now set..

    #-----------------------------------
    # Call processes and store in hash..
    #-----------------------------------
    $iProcessCount = Proc::GetLongProcessListHash ( \@hProcListHash );

    #----------
    # Restore..
    #----------
    defined $sLang ? $ENV { LANG } = $sLang:delete $ENV { LANG }; # Restore..

    #----------------------
    # Check if we got any..
    #----------------------
    if ( $iProcessCount ) {
      &logTrace ( 20075,
      "<%s> Got [%s] processes from proc table for filtering",
      $sMyName, $iProcessCount );
    }
    else {
      &logError ( 20076,
      "<%s> Call to get process list did not return any processes",
      $sMyName );
      return VCSDef_MonitorExit::OFFLINE; # Offline..
    }

    #-----------------------
    # Search for processes..
    #-----------------------
    foreach $rhlProc ( @hProcListHash ) {
      #--------------------------------------------------------------
      # Each $rhlProc is a ref. to a hash with the following keys..
      # "UID"     =>  <user id>
      # "PID"     =>  <process id>
      # "PPID"    =>  <parent process id>
      # "COMMAND" =>  <command string>
      #--------------------------------------------------------------

      &logTrace ( 20077, "<%s> Searching PID [%s] UID [%s] CMD [%s]",
      $sMyName, $$rhlProc { PID }, $$rhlProc { UID },
      $$rhlProc { COMMAND } );

      next unless defined $$rhlProc { UID }; # Bug in Proc::GetProcessListHash..
      $iPid = $$rhlProc { PID };             # Assign..

      if ( $$rhlProc { UID } == $iUid ) { # Process owner..
        #-----------------------------------------------
        # Look for processes that were offline in FFLM..
        #-----------------------------------------------
        if ( $iUseFFLM != 2 ) { # Request to filter..
          foreach $sKey ( keys %hsCopy ) {
            @lsUpdates = (); # Reset..
            @lsValues = @{ $hsCopy { $sKey } }; # Get values..
    
            #--------------------------------------------------
            # Should have details per three elements of array..
            # <PIDFILE> <PROCESS> <STATE>
            #--------------------------------------------------
            while ( ( $sPidFile, $sProc, $iFound ) =
                      splice ( @lsValues, 0, 3 ) ) {
              if ( ( VCSDef_MonitorExit::OFFLINE == $iFound ) && # Earlier marked offline..
                   ( $$rhlProc { COMMAND } =~ /$sProc/ ) &&      # Now found..
                   ( ! defined $hiLCPid { $iPid } ) ) {          # Not already matched..

                $hiLCPid { $iPid } = $$rhlProc { PPID };         # Store PPID of the process.. 
    
                &logTrace ( 20078,
                "<%s> Found PID [%s] PPID [%s] CMD [%s] with filter [%s]",
                $sMyName, $iPid, $$rhlProc { PPID }, $$rhlProc { COMMAND }, $sProc );

                $hiProcPids { $sKey } = $iPid;        # Store..
                unshift ( @$rliOnlinePids, $iPid );   # Store..
                $iFound = VCSDef_MonitorExit::ONLINE; # Mark online..

                &logTrace ( 20079,
                "<%s> Got [%s] processes into hiLCPid", $sMyName, scalar(keys(%hiLCPid)) );

                #----------------
                # Update status..
                #----------------
                push ( @lsUpdates, $sPidFile, $sProc, $iFound );
                @{ $hsCopy { $sKey } } = @lsUpdates; # Update..

                last;
              }
            }
          }
        }
        else { # Request for kill..
          &logTrace ( 20080,
          "<%s> Pushing PID [%s] with PPID [%s] Cmd [%s] on stack",
          $sMyName, $iPid, $$rhlProc { PPID }, $$rhlProc { COMMAND } );
          push ( @$rliOnlinePids, $iPid ); # Store..

          $hiLCPid { $iPid } = $$rhlProc { PPID };         # Store PPID of the process..
        } 
      }
    }
  }

  #------------------------------------
  # Set final status for each process..
  #------------------------------------
  my ( $iOnline, $iOffline, $sEntryPointName ) = ();

  ( $sEntryPointName = $0 ) =~ s#.*/##; # Basename..
  $sTemp = ''; # Ensure..

  foreach $sKey ( keys %hsCopy ) {
    @lsValues = @{ $hsCopy { $sKey } }; # Get values..
    $sTemp .= "\nProcess      [$sKey]\n";

    while ( ( $sPidFile, $sProc, $iFound ) =
              splice ( @lsValues, 0, 3 ) ) {
      #-----------------------------------
      # Create PID files for use in FFLM..
      #-----------------------------------
      if ( $iUseRFLM                                 &&      # In RFLM..
           $iUseFFLM != 2                            &&      # Asked to filter..
           $hiProcPids { $sKey }                     &&      # PID populated..
           ( $sEntryPointName =~ /^(monitor|online)$/o ) ) { # In monitor or online..

        &logTrace ( 20081,
        "<%s> Creating PID file [%s] with PID [%s]",
        $sMyName, $sPidFile, $hiProcPids { $sKey } );

        #---------
        # Create..
        #---------
        if ( SAPLiveCache::CreatePidFile ( $sPidFile,
                                           $hiProcPids { $sKey } ) ) {
          &logTrace ( 20082,
          "<%s> Successfully created PID file [%s]", $sMyName, $sPidFile );
        }
        else {
          &logError ( 20083,
          "<%s> Could not create PID file [%s] successfully",
          $sMyName, $sPidFile );
        }
      }

      #-------------
      # Set status..
      #-------------
      if ( VCSDef_MonitorExit::ONLINE == $iFound && ( 2 <= scalar (keys (%hiLCPid)) ) ) {
        $iOnline = $iFound;
      }
      elsif ( VCSDef_MonitorExit::OFFLINE == $iFound ) {
        $iOffline = $iFound;
      }

      $sTemp .= "Srch Process [$sProc]\n";
      $sTemp .= "FLM Status   [$iFound]\n\n";
    }

    $sTemp .= "--------------------";
  }

  #-------------------
  # Set final status..
  #-------------------
  $iReturnCode = $iOffline || $iOnline || VCSDef_MonitorExit::UNKNOWN;

  #------------------------------------------------
  # Reset status based on PIDs, if not in monitor..
  #------------------------------------------------
  if ( $sEntryPointName ne 'monitor'                   &&
       ( VCSDef_MonitorExit::OFFLINE == $iReturnCode ) &&
       scalar @$rliOnlinePids ) {
    $iReturnCode = VCSDef_MonitorExit::PARTIAL_ONLINE;

    &logTrace ( 20084,
    "<%s> Found [%s] PIDs [%s] for this instance of SAPLiveCache.
    Resetting return code to [%s]",
    $sMyName, scalar @$rliOnlinePids, join ( ' ', @$rliOnlinePids ),
    $iReturnCode );
  }

  #-----------------------
  # Report on FLM status..
  #-----------------------
  if ( $iUseFFLM != 2 ) { # Request to filter..
    &logTrace ( 20085,
    "<%s> SAPLiveCache process details ---\n%s\nReturn code is [%s]",
    $sMyName, $sTemp, $iReturnCode );
    $sTemp = undef; # Release..
  }

  # Sort the PIDs of the liveCache processes in the descending order of their parent PIDs.
  if( (scalar @$rliOnlinePids > 0) && (scalar @$rliOnlinePids == scalar (keys (%hiLCPid)) ) ) {
    &logTrace ( 20086,
    "<%s> Sorting the liveCache process IDs in the descending order of their parent IDs.",
    $sMyName );

    @$rliOnlinePids = sort {$hiLCPid{$b} <=> $hiLCPid{$a} } keys %hiLCPid ;
  }

  return $iReturnCode;
}



=pod

 -------------------------------------------------------------------------
 Function : int Monitor ( IN String DBMPassword,
                          IN String DBMUserId,
                          IN String EnvFile,
                          IN String LCHome,
                          IN String LCName,
                          IN String LCUser,
                          IN String MonitorProgram,
                          IN Integer SecondLevelMonitor );

 Description: Performs the in-depth monitoring

 Parameters :

 $sDBMPassword              Password for the user <DBMUserId>.
 $sDBMUserId                User ID used to connect to liveCache server.
 $sEnvFile                  Environments file for <LCUser>. 
 $sLCHome                   Directory for LiveCache database software.
 $sLCName                   Name of the liveCache instance.
 $sLCUser                   Operating system user for liveCache server.
 $sMonitorProgram           Full path to external monitor program.
 $iSecondLevelMonitor       Cycle count to run in-depth monitoring.
 
 Returns : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 110  ONLINE with confidence level 100          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 -------------------------------------------------------------------------

=cut

sub Monitor {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sDBMPassword,
       $sDBMUserId,
       $sEnvFile,
       $sLCHome,
       $sLCName,
       $sLCUser,
       $sMonitorProgram,
       $iSecondLevelMonitor ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20087, "Subroutine <%s> called with:
  DBMPassword              [%s]
  DBMUserId                [%s]
  EnvFile                  [%s]
  LCHome                   [%s]
  LCName                   [%s]
  LCUser                   [%s]
  MonitorProgram           [%s]
  SecondLevelMonitor       [%s]",

  $sMyName,  $sDBMPassword,   $sDBMUserId,   $sEnvFile,   $sLCHome,
  $sLCName, $sLCUser, $sMonitorProgram, $iSecondLevelMonitor );

  #-----------------------------
  # Set return code to success..
  #-----------------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE; # Success..

  my $iTillWhen = undef;

  #----------------------------------------------
  # In-depth monitoring, if correct time to run..
  #----------------------------------------------
  if ( VCSagentFW::RunSecondLevelMonitor ( $iSecondLevelMonitor ) ) {
    #--------------------------
    # Just In Time Validation..
    #--------------------------
    if ( VCSDef_MonitorExit::ONLINE == ( $iReturnCode =
         SAPLiveCache::DelayedAttributeCheck ( $sDBMPassword,
                                               $sDBMUserId,
                                               $sEnvFile,
                                               $sLCHome,
                                               $sLCName,
                                               $sLCUser ) ) ) {
      &logTrace ( 20088,
      "<%s> Delayed attribute check completed successfully",
      $sMyName );
    }
    else {
      &logError ( 20089,
      "<%s> Attribute validation failed. Returning [%s]",
      $sMyName, $iReturnCode );
      return $iReturnCode;
    }

    &logTrace ( 20090,
    "<%s> Proceeding with second level monitoring", $sMyName );

    #-------------------------------------------------------
    # Upper bound..
    # Time in EPOCH when entry pt started + VCS time limit..
    #-------------------------------------------------------
    $iTillWhen = $^T + SAPLiveCache::GetEngineTimeout ( 1 );

    if ( VCSDef_MonitorExit::ONLINE == ( $iReturnCode =
         SAPLiveCache::SecondLevelMonitor ( $sDBMPassword,
                                            $sDBMUserId,
                                            $sEnvFile,
                                            $sLCHome,
                                            $sLCName,
                                            $sLCUser,
                                            $iTillWhen ) ) ) {

      &logTrace ( 20091,
      "<%s> SAPLiveCache instance online with status [%s] after SLM",
      $sMyName, $iReturnCode );
    }
    else {
      &logError ( 20092,
      "<%s> SAPLiveCache instance NOT online after SLM. Status [%s]",
      $sMyName, $iReturnCode );
      return $iReturnCode;
    }
  }

  #------------------------------------------------
  # Exit if external monitor has not been defined..
  #------------------------------------------------
  return $iReturnCode unless defined $sMonitorProgram;

  #---------------------------------------
  # MonitorProgram defined..
  # Separate the executable and the args..
  #---------------------------------------
  my ( $sExtCmd, $sExtCmdArgs ) = $sMonitorProgram # Separate
                                =~ /(\S+)\s*(.*)/; # Cmd and args..

  $sExtCmdArgs = '' unless defined $sExtCmdArgs;   # Prevent undefs..

  if ( $sExtCmd ) {                                # Defined..
    #----------------------------------------
    # Get time bearings, if not got earlier..
    #----------------------------------------
    $iTillWhen ||= $^T + SAPLiveCache::GetEngineTimeout ( 1 );

    if ( VCSDef_MonitorExit::ONLINE ==             # and found online..
         ( $iReturnCode =                          # by the status..
         SAPLiveCache::ExternalMonitor ( $sLCUser,
                                         $sEnvFile,
                                         $sExtCmd,
                                         $sExtCmdArgs,
                                         $iTillWhen ) ) ) {
      &logTrace ( 20093,
      "<%s> SAPLiveCache instance found online with status [%s]
      after executing external monitor [%s] with args [%s]",
      $sMyName, $iReturnCode, $sExtCmd, $sExtCmdArgs );
    }
    else {
      &logError ( 20094,
      "<%s> SAPLiveCache instance NOT online after executing external monitor
      [%s] with args [%s]. Status [%s]",
      $sMyName, $sExtCmd, $sExtCmdArgs, $iReturnCode );
    }
  }

  return $iReturnCode;
}

=pod

 -------------------------------------------------------------------------
 Function : int DelayedAttributeCheck ( IN String DBMPassword,
                                        IN String DBMUserId,
                                        IN String EnvFile,
                                        IN String LCHome,
                                        IN String LCName,
                                        IN String LCUser );

 Description : Just In Time Validation

 Parameters :

 $sDBMPassword              Password for the user <DBMUserId>.
 $sDBMUserId                User ID used to connect to liveCache server.
 $sEnvFile                  Environments file for <LCUser>. 
 $sLCHome                   Directory for LiveCache database software.
 $sLCName                   Name of the liveCache instance.
 $sLCUser                   Operating system user for liveCache server.
 
 Returns  : Integer value as defined below:

 110  ONLINE with confidence level 110          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 -------------------------------------------------------------------------

=cut

sub DelayedAttributeCheck {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sDBMPassword,
       $sDBMUserId,
       $sEnvFile,
       $sLCHome,
       $sLCName,
       $sLCUser ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20095, "Subroutine <%s> called with:
  DBMPassword              [%s]
  DBMUserId                [%s]
  EnvFile                  [%s]
  LCHome                   [%s]
  LCName                   [%s]
  LCUser                   [%s]",

  $sMyName,  $sDBMPassword,   $sDBMUserId,   $sEnvFile,   $sLCHome,
  $sLCName, $sLCUser );

  #-----------------------------
  # Set return code to success..
  #-----------------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE; # Success..

  #------------------------------
  # Validate parameter EnvFile..
  #------------------------------
  if ( ! -f  $sEnvFile ) {
    &logWarn ( 20096, "<%s> [%s] is not a plain file",
    $sMyName, $sEnvFile );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Failure..
  }

  #-----------------------------
  # Validate parameter LCHome..
  #-----------------------------
  if ( ! -d  $sLCHome ) {
    &logWarn ( 20097, "<%s> [%s] directory does not exist",
    $sMyName, $sLCHome );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Failure..
  }

  #-----------------------------
  # Validate parameter LCUser..
  #-----------------------------
  if ( ! defined getpwnam $sLCUser ) {
    &logWarn ( 20098, "<%s> [%s] is not a valid UNIX login",
    $sMyName, $sLCUser );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Failure..
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : int SecondLevelMonitor ( IN String DBMPassword,
                                     IN String DBMUserId,
                                     IN String EnvFile,
                                     IN String LCHome,
                                     IN String LCName,
                                     IN String LCUser,
                                     IN Integer TillWhen );

 Description : Does in-depth monitoring, if enabled.

 Parameters :

 $sDBMPassword              Password for the user <DBMUserId>.
 $sDBMUserId                User ID used to connect to liveCache server.
 $sEnvFile                  Environments file for <LCUser>. 
 $sLCHome                   Directory for LiveCache database software.
 $sLCName                   Name of the liveCache instance.
 $sLCUser                   Operating system user for liveCache server.
 $iTillWhen                 Time in EPOCH left for IInd level monitoring
 
 Returns  : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 110  ONLINE with confidence level 110          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 -------------------------------------------------------------------------

=cut

sub SecondLevelMonitor {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sDBMPassword,
       $sDBMUserId,
       $sEnvFile,
       $sLCHome,
       $sLCName,
       $sLCUser,
       $iTillWhen ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20099, "Subroutine <%s> called with:
  DBMPassword              [%s]
  DBMUserId                [%s]
  EnvFile                  [%s]
  LCHome                   [%s]
  LCName                   [%s]
  LCUser                   [%s]
  TillWhen                 [%s]",

  $sMyName,  $sDBMPassword,   $sDBMUserId,   $sEnvFile,   $sLCHome,
  $sLCName, $sLCUser, $iTillWhen );  

  #------------------
  # Hopeful success..
  #------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE;

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $iTimeout, $iExitCode, $sCommand, $sCommandArgs, @lsCmdout ) = ();

  #---------------------------------------------
  # Declare monitor command and args variables..
  #---------------------------------------------
  $sCommand = "$sLCHome/dbmcli";
  $sCommandArgs = "-d $sLCName -u $sDBMUserId,$sDBMPassword db_state";

  #-------------------------------
  # Validate and execute command..
  #-------------------------------
  if ( ( $iTimeout = $iTillWhen - time ) <= 10 ) { # Get time bearings..
    &logError ( 20100,
    "<%s> Insufficient time to continue in-depth monitoring.
    Short by [%s] sec",
    $sMyName, abs ( $iTimeout ) );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Failure..
  }
  elsif ( SAPLiveCache::ExecLocalCmd ( $sLCUser, 
                                       $sEnvFile, 
                                       $sCommand, 
                                       $sCommandArgs, 
                                       \$iExitCode, 
                                       $iTimeout, 
                                       \@lsCmdout, 
                                       $sDBMPassword ) ) {
    &logTrace ( 20101,
    "<%s> Cmd [%s] with args [%s] executed successfully",
    $sMyName, $sCommand, $sCommandArgs );

    if ( $lsCmdout [-1] eq 'ONLINE' ) {
      &logTrace ( 20102,
      "<%s> Found search pattern [%s] in the output [%s].",  
      $sMyName, $lsCmdout [-1], join ( chr 012, @lsCmdout ) ); 
    }
    elsif ( $lsCmdout [-1] eq 'ADMIN' ) {
      &logInfo ( 20103,
      "<%s> Found search pattern [%s] in the output [%s].",  
      $sMyName, $lsCmdout [-1], join ( chr 012, @lsCmdout ) ); 
    }
    elsif ( $lsCmdout [-1] eq 'STARTING' ) {
      &logInfo ( 20104,
      "<%s> Found search pattern [%s] in the output [%s].",  
      $sMyName, $lsCmdout [-1], join ( chr 012, @lsCmdout ) ); 
    }
    elsif ( $lsCmdout [-1] eq 'OFFLINE' ) {
      &logError ( 20105,
      "<%s> Found search pattern [%s] in the output [%s].",  
      $sMyName, $lsCmdout [-1], join ( chr 012, @lsCmdout ) ); 

      $iReturnCode = VCSDef_MonitorExit::OFFLINE;
    }
    elsif ( $lsCmdout [-1] eq 'SHUTDOWN' ) {
      &logError ( 20106,
      "<%s> Found search pattern [%s] in the output [%s].",  
      $sMyName, $lsCmdout [-1], join ( chr 012, @lsCmdout ) ); 

      $iReturnCode = VCSDef_MonitorExit::OFFLINE; 
    }
    else {
      &logError ( 20107,
      "<%s> SecondLevelMonitor returned undefined output [%s].", 
      $sMyName, join ( chr 012, @lsCmdout ) ); 

      $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
    }
  }
  elsif ( ! defined $iExitCode ) {
    &logError ( 20108,
    "<%s> Second Level Monitor returned undefined exit code and Cmd output [%s]",
    $sMyName, join ( chr 012, @lsCmdout ) );

    $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Failure..
  }
  else {
    &logError ( 20109,
    "<%s> Second Level Monitor returned invalid exit code [%s] and Cmd output [%s]",
    $sMyName, $iExitCode, join ( chr 012, @lsCmdout ) );

    $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Failure..
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : int ExternalMonitor ( IN String LCUser,
                                  IN String EnvFile,
                                  IN String Cmd,
                                  IN String CmdArgs,
                                  IN Integer TillWhen );

 Description : Handles user defined external monitor program

 Parameters :

 $sLCUser                   Operating system user for liveCache server.
 $sEnvFile                  Environments file for <LCUser>. 
 $sCmd                      User defined utilty to execute
 $sCmdArgs                  Command line arguments
 $iTillWhen                 Time in EPOCH left for IInd level monitoring
 
 Returns  : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 110  ONLINE with confidence level 110          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 -------------------------------------------------------------------------

=cut

sub ExternalMonitor {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sLCUser,
       $sEnvFile,
       $sCmd,
       $sCmdArgs,
       $iTillWhen ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20110, "Subroutine <%s> called with:
  LCUser                   [%s]
  EnvFile                  [%s]
  Cmd                      [%s]
  CmdArgs                  [%s]
  TillWhen                 [%s]",

  $sMyName, $sLCUser, $sEnvFile, $sCmd, $sCmdArgs, $iTillWhen );

  #------------------
  # Hopeful success..
  #------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE;

  #------------------------
  # Setup local variables..
  #------------------------
  my ( $iExitCode, $iTimeout, @lsCmdOut ) = ();
  my $iBuf = 5; # Secs..

  #-------------------------
  # Handled exit code hash..
  #-------------------------
  my %hiExitCodes = (  0, VCSDef_MonitorExit::ONLINE,
                     110, VCSDef_MonitorExit::ONLINE,
                       1, VCSDef_MonitorExit::OFFLINE,
                     100, VCSDef_MonitorExit::OFFLINE );

  #---------------------------------------------
  # Check if command can be found and executed..
  #---------------------------------------------
  if ( ! -e $sCmd ) {
   &logError ( 20111,
    "<%s> External monitor [%s] does not exist",
    $sMyName, $sCmd );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
  }
  elsif ( ! -x $sCmd ) {
   &logError ( 20112,
    "<%s> External monitor [%s] not executable.",
    $sMyName, $sCmd );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
  }
  elsif ( ( $iTimeout = $iTillWhen - ( time + $iBuf ) ) <= 5 ) { # Lack of time..
   &logError ( 20113,
    "<%s> Insufficient time to proceed with external monitor [%s].
    Short by [%s] sec", $sMyName, $sCmd, abs ( $iTimeout ) );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
  }

  #--------------------------------------------
  # Return if validation failed at this point..
  #--------------------------------------------
  unless ( VCSDef_MonitorExit::ONLINE == $iReturnCode ) {
    return $iReturnCode;
  }

  &logTrace ( 20114,
  "<%s> Executing external monitor with Command [%s] Args [%s] in [%s] secs",
  $sMyName, $sCmd, $sCmdArgs, $iTimeout );

  #----------------------
  # Execute the command..
  #----------------------
  SAPLiveCache::ExecLocalCmd ( $sLCUser, 
                               $sEnvFile, 
                               $sCmd, 
                               $sCmdArgs, 
                               \$iExitCode, 
                               $iTimeout, 
                               \@lsCmdOut );
                               
  #------------------------
  # Report on exit status..
  #------------------------
  if ( ! defined $iExitCode ) {
    &logError ( 20115,
    "<%s> Command [%s] with args [%s] timed out",
    $sMyName, $sCmd, $sCmdArgs );
    $iExitCode = 'undef';
  }
  elsif ( $hiExitCodes { $iExitCode } &&
          VCSDef_MonitorExit::ONLINE ==
          $hiExitCodes { $iExitCode } ) {
    &logTrace ( 20116,
    "<%s> Command [%s] with args [%s] executed successfully with exit code [%s].
    Cmd output [%s]",
    $sMyName, $sCmd, $sCmdArgs, $iExitCode, join ( chr 012, @lsCmdOut ) );
  }
  else {
    &logWarn ( 20117,
    "<%s> Command [%s] with args [%s] execution failed with exit code [%s].
    Cmd output [%s]",
    $sMyName, $sCmd, $sCmdArgs, $iExitCode, join ( chr 012, @lsCmdOut ) );
  }

  #--------------------------------
  # Setup appropriate return code..
  #--------------------------------
  $iReturnCode = $hiExitCodes { $iExitCode } ||
                 VCSDef_MonitorExit::UNKNOWN;

  return $iReturnCode;
}



=pod

 -------------------------------------------------------------------------
 Function : bool CreatePidFile ( IN String PidFile,
                                 In Integer Pid );


 Description : Creates PID file with given PID

 Parameters :

 $sPidFile     File to create
 $iPid         PID to store

 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub CreatePidFile {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sPidFile,
       $iPid ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20118, "Subroutine <%s> called with:
  PidFile         [%s]
  Pid             [%s]",

  $sMyName, $sPidFile, $iPid );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1;

  #-----------------------
  # Check against undefs..
  #-----------------------
  unless ( defined $iPid && $iPid =~ /^\d+$/o) {
    $iPid ||= 'undef'; # Prevent undefs.
    &logError ( 20119, "<%s> PID [%s] is not a valid positive number",
    $sMyName, $iPid );
    return 0; # Proceed no further..
  }

  #------------------
  # Create PID file..
  #------------------
  &logTrace ( 20120,
  "<%s> Creating PID file [%s] with PID [%s]",
  $sMyName, $sPidFile, $iPid );

  if ( sysopen ( PIDFILE, $sPidFile, O_WRONLY|O_SYNC|O_CREAT|O_TRUNC,
                 0600 ) ) {
    print PIDFILE $iPid, "\n"; # Update..

    unless ( close ( PIDFILE ) ) {
      &logWarn ( 20121,
      "<%s> Failed to close PidFile [%s] after update: [%s]",
      $sMyName, $sPidFile, $! );
      $iReturnCode = 0;
    }
  }
  else {
    &logWarn ( 20122,
    "<%s> Failed to create PID file [%s] with PID [%s]",
    $sMyName, $sPidFile, $iPid );
    $iReturnCode = 0;
  }

  return $iReturnCode;
}



=pod

 -------------------------------------------------------------------------
 Function : int FFLMStatus ( IN String PidFile,
                             IN String Proc,
                             IN Integer Uid,
                             OUT RefInt Pid );

 Description: Checks status of Process listed in PidFile, if running

 Parameters :

 $sPidFile       Absolute path to PID file of service
 $sProc          Process string to search for in proc table
 $iUid           UNIX UID of owner
 $riPid          Reference to existing PID

 Returns : Integer value as defined below:
 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 110  ONLINE with confidence level 100          VCSDef_MonitorExit::ONLINE
 -------------------------------------------------------------------------

=cut

sub FFLMStatus {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sPidFile,
       $sProc,
       $iUid,
       $riPid,
       $riPPid ) = @_;  # by vchappav ...

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20123, "Subroutine <%s> called with:
  PidFile                [%s]
  Proc                   [%s]
  UID                    [%s]
  riPid                  [%s]",

  $sMyName, $sPidFile, $sProc, $iUid, $riPid );

  #-----------------------------
  # Set return code to success..
  #-----------------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE; # Success..
  $$riPid = 0; # Reset..

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $iPid, $rhlProc, @lProcessHash ) = ();

  #----------------------------------
  # Start by testing the attributes..
  #----------------------------------
  if ( $sProc !~ /\S/ ) {
    &logError ( 20124, "<%s> Called with empty [%s] process string",
    $sMyName, $sProc );
    $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
  }
  elsif ( ! -e $sPidFile ) {
    &logTrace ( 20125, "<%s> PID file [%s] does not exist",
    $sMyName, $sPidFile );
    $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
  }
  elsif ( ! -s $sPidFile ) {
    &logWarn ( 20126, "<%s> PID file [%s] is of zero bytes",
    $sMyName, $sPidFile );
    $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
  }

  #---------------------
  # Return, if offline..
  #---------------------
  return $iReturnCode unless VCSDef_MonitorExit::ONLINE == $iReturnCode;
 
  if ( open ( PIDFILE, $sPidFile ) ) {
    ( $iPid ) = grep { ( s/\n// || 1 ) && # Chomp..
                         s/.*?(\d+)$/$1/  # Get PID..
                } <PIDFILE>;

    $iPid = '' unless defined $iPid;      # Avoid undefs..

    close ( PIDFILE ) ||                  # Close file..
    &logWarn ( 20127,
    "<%s> Could not close PID file [%s] after read [%s]",
    $sMyName, $sPidFile, $! );
  }
  else {
    &logError ( 20128,
    "<%s> Failed to open PID file [%s] for read: [%s]",
    $sMyName, $sPidFile, $! );
    return VCSDef_MonitorExit::OFFLINE; # Failure..
  }

  #------------------------
  # Check if PID is valid..
  #------------------------
  if ( $iPid =~ /\d+/ &&     # If the valid number..
       kill ( 0, $iPid ) ) { # is a PID that's alive..

    &logTrace ( 20129,
    "<%s> PID [%s] from PID file [%s] exists",
    $sMyName, $iPid, $sPidFile );

    #-----------------------
    # Search in proc table..
    #-----------------------
    if ( SAPLiveCache::GetPidDetails ( $iPid,
                                       \@lProcessHash ) ) {
      #--------------
      # Got details..
      #--------------
      foreach $rhlProc ( @lProcessHash ) {
        &logTrace ( 20130,
        "<%s> Process details: UID [%s] PID [%s] PPID [%s] ARGS [%s]",
        $sMyName, $$rhlProc { UID }, $$rhlProc { PID }, $$rhlProc { PPID },
        $$rhlProc { COMMAND } );

        #--------------
        # Now compare..
        #--------------
        if ( ( $$rhlProc { UID } == $iUid ) &&       # Owner match..
               $$rhlProc { COMMAND } =~ /$sProc/ ) { # Filter match..
          &logTrace ( 20131,
          "<%s> PID [%s] owned by UID [%s] with ARGS [%s] contains [%s]",
          $sMyName, $iPid, $$rhlProc { UID }, $$rhlProc { COMMAND }, $sProc );
          $$riPid  = $iPid; # Store..
          $$riPPid = $$rhlProc { PPID }; # Store..  # by vchappav ...
        }
        else {
          &logTrace ( 20132,
          "<%s> Ignoring PID [%s] owned by UID [%s] with ARGS [%s] with string [%s]",
          $sMyName, $iPid, $$rhlProc { UID }, $$rhlProc { COMMAND }, $sProc );
          $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
        }
      }
    }
    else {
      &logTrace ( 20133,
      "<%s> Could not get details for PID [%s]",
      $sMyName, $iPid );
      $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
    }
  }
  else {
    &logTrace ( 20134, "<%s> PID [$iPid] does not exist",
    $sMyName, $sPidFile );
    $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
  }

  @lProcessHash = (); # Truncate..
  return $iReturnCode;
}



=pod

 -------------------------------------------------------------------------
 Function : bool KillSAPLiveCacheProcesses ( IN String LCName,
                                             IN String LCUser );

 Description  : Kills processes for this instance of SAPLiveCache
 
 Parameters :

 $sLCName                 Name of the liveCache instance.
 $sLCUser                 Operating system user for liveCache server.
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub KillSAPLiveCacheProcesses {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sLCName,
       $sLCUser ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace( 20135, "Subroutine <%s> called with:
  LCName                   [%s]
  LCUser                   [%s]",

  $sMyName, $sLCName, $sLCUser );
 
  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1; # Success..

  #-------------------------
  # Set up local variables..
  #-------------------------
  my ( @liGoners, @liUnDead ) = ();

  #---------------------------------------------
  # Get list of processes running for this Uid..
  #---------------------------------------------
  if ( ( caller 1 )[3] eq 'SAPLiveCache::Start' ) { # During online..
    @liGoners = @liRunningProcs;              # Get PIDs to kill..
  }
  else {
    SAPLiveCache::GetProcState ( \@liGoners,
                                 $sLCName,
                                 $sLCUser,
                                 0 ); # Set filtering choice..
  }

  #--------------------------
  # Sort in ascending order..
  #--------------------------
  #@liGoners = sort { $a <=> $b } @liGoners;        # by vchappav ...

  #------------------------------------------------------------------
  # We do a two level kill. The SIGINQ ensures  that  the  PID  is  a
  # valid process before we actually issue the SIGKILL. We don't want
  # to report a failure for SIGKILL in case  of  defunct  or  missing
  # processes.
  #------------------------------------------------------------------
  if ( scalar @liGoners ) {
    #----------------------
    # Failed to kill some..
    #----------------------
    my $iPid = ();
    foreach $iPid ( @liGoners ) {
      if ( kill ( 0, $iPid ) ) {
        if( ! kill ( 9, $iPid ) ) { # could NOT be killed..
          &logError ( 20136,
          "<%s> PID not killed [%s].",
          $sMyName, $iPid ); 

          $iReturnCode = 0; # Failure..
        }
      }
      else { # killed..
        &logInfo ( 20137,
        "<%s> PID [%s] doesn't exist for killing.", $sMyName, $iPid ); 
      }
    }
  }
  else {
    &logTrace ( 20138,
    "<%s> Could not locate any processes for this instance of SAPLiveCache",
    $sMyName );
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool ExecLocalCmd ( IN String User,
                                IN String EnvFile,
                                IN String Command,
                                IN String CommandArgs,
                                OUT RefInt ExitCode,
                                IN Integer Timeout,
                                OUT ListReference CmdOut ,
                                IN String HideString );

 Description : Wrapper function for
               Sys::RunWithEnvCmdWithOutputWithTimeOut to collect output
               of the command.
 
 Parameters :

 $sUser            UNIX user to run the command as
 $sEnvFile         Absolute path to location of EnvFile to source
 $sCommand         Absolute path to location of command
 $sCommandArgs     Arguments to command
 $riExitCode       Reference to collect exit code of command
 $iTimeout         Time limit to execute the command
 $rlsCmdOut        Stores o/p from command if successful
 $sHideString      String in CommandArgs to hide

 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub ExecLocalCmd {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sUser,
       $sEnvFile,
       $sCommand,
       $sCommandArgs,
       $riExitCode,
       $iTimeout,
       $rlsCmdOut ,
       $sHideString ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace( 20139, "Subroutine <%s> called with:
  User        [%s]
  EnvFile     [%s]
  Command     [%s]
  CommandArgs [%s]
  riExitCode  [%s]
  Timeout     [%s]
  rlsCmdOut   [%s]",

  $sMyName,
  $sUser, $sEnvFile, $sCommand, $sCommandArgs, $riExitCode,
  $iTimeout, $rlsCmdOut );

  #---------------------------------------
  # Setup return code to hopeful success..
  #---------------------------------------
  my $iReturnCode = 1;

  my ( $iExitCode, $sDispStr, $sCmdDeCr, $sDeCrPasswd ) = ();

  #---------------------
  # Set up for display..
  #---------------------
  $sDispStr = $sCommandArgs;

  #-------------------------
  # Get decrypted password..
  #-------------------------
  if ( defined $sHideString ) {
    #-----------
    # Validate..
    #-----------
    if ( ! -x $sCmdVcsDecrypt ) {
      &logError ( 20140,
      "<%s> Command [%s] is not executable", $sMyName, $sCmdVcsDecrypt );
      return 0; # Failure..
    }

    $sCmdDeCr = $sCmdVcsDecrypt;                          # Get global..
    ( $sDeCrPasswd ) = grep {
                         s/([^\w])/\\$1/g || 1
                       } readpipe ( "$sCmdDeCr $sHideString" );

    #--------------------
    # Collect exit code..
    #--------------------
    if ( $iExitCode = $? >> 8 ) {
      &logError ( 20141, "<%s> Failed to decrypt crypted string [%s]",
      $sMyName, $sHideString );
      return 0; # Failure..
    }

    $sCommandArgs =~ s/$sHideString/$sDeCrPasswd/g;
  }

  #-------------------
  # Set up arguments..
  #-------------------
  my $bSourceUserEnv = 0; # Do not source..
  my $bTimedOut = 0;      # Timeout variable..

  @$rlsCmdOut = (); # Reset..

  #-----------
  # Validate..
  #-----------
  if ( ! -e $sCommand ) {
    &logError ( 20142,
    "<%s> Cmd [%s] does not exist", $sMyName, $sCommand );
    $iReturnCode = 0; # Failure..
  }
  elsif ( ! -x $sCommand ) {
    &logError ( 20143,
    "<%s> Cmd [%s] is not executable", $sMyName, $sCommand );
    $iReturnCode = 0; # Failure..
  }
  elsif ( $iTimeout <= 1 ) {
    &logError ( 20144,
    "<%s> Insufficient time [%s] to execute Cmd [%s] with Args [%s].",
    $sMyName, $iTimeout, $sCommand, $sDispStr );
    $iReturnCode = 0; # Failure..
  }

  $iReturnCode || return $iReturnCode; # Return if failed.. 

  #----------------------
  # Set LOGNAME to user..
  #----------------------
  my $sOldLogName = $ENV { LOGNAME }; # Backup..
  $ENV { LOGNAME } = $sUser;          # Now set..

  &logTrace ( 20145,
  "<%s> Executing Command [%s] with Args [%s]",
  $sMyName, $sCommand, $sDispStr );

  $iReturnCode = Sys::RunWithEnvCmdWithOutputWithTimeOut ( $sEnvFile,
                                                           $bSourceUserEnv,
                                                           $sCommand,
                                                           $sCommandArgs,
                                                           $riExitCode,
                                                           $sUser,
                                                           $rlsCmdOut,
                                                           \$bTimedOut,
                                                           $iTimeout,
                                                           undef,
                                                           undef,
                                                           $sDeCrPasswd );

  #---------------------------
  # Remove trailing newlines..
  #---------------------------
  chomp @$rlsCmdOut;

  #--------------------
  # Check return code..
  #--------------------
  if ( $iReturnCode ) {   # No error with command execution..
    if ( $$riExitCode ) { # Failure..
      &logTrace ( 20146,
      "<%s> Command [%s] with args [%s] returned exit code [%s]",
      $sMyName, $sCommand, $sDispStr, $$riExitCode );
      $iReturnCode = 0; # Failure..
    }
    else {
      &logTrace ( 20147,
      "<%s> Command [%s] with args [%s] executed successfully with exit code [%s]",
      $sMyName, $sCommand, $sDispStr, $$riExitCode );
    }
  }
  else { # Execution error..
    if ( $bTimedOut ) {  # We timed out..
      &logError ( 20148,
      "<%s> Command [%s] with args [%s] did not complete within [%s] seconds",
      $sMyName, $sCommand, $sDispStr, $iTimeout );
      $iReturnCode = 0; # Failure..
    }
    else { # Failure due to something else..
      &logError ( 20149,
      "<%s> Failed in call to RunWithEnvCmdWithOutputWithTimeOut",
      $sMyName );
      $iReturnCode = 0; # Failure..
    }
    @$rlsCmdOut = (); # Reset..
  }

  #----------
  # Restore..
  #----------
  defined $sOldLogName ? $ENV { LOGNAME } = $sOldLogName:delete $ENV { LOGNAME };

  return $iReturnCode;
}



=pod

 -------------------------------------------------------------------------
 Function : bool GetPidDetails ( IN Integer Pid,
                                 OUT RefList ProcList );


 Description : Gets long process details for a known Pid if running

 Parameters :

 $iPid         PID whose details are requested
 $rlProcList   Ref to List to store details
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub GetPidDetails {
  #----------------
  # Get Arguments..
  #----------------
  my ( $iPid,
       $rlProcList ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20150, "Subroutine <%s> called with:
  Pid             [%s]
  rlProcList      [%s]",

  $sMyName, $iPid, $rlProcList );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1;

  #---------------
  # Validate PID..
  #---------------
  unless ( defined $iPid    &&   # Catch undefs..
           $iPid =~ /^\d+$/ &&   # And invalid PIDs..
           kill ( 0, $iPid ) ) { # And defunct ones..
    &logError ( 20151,
    "<%s> PID [%s] is invalid", $sMyName, $iPid );
    return 0;
  }

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $sPsCmd, $sPsCmdArgs, $sLang, $sFmt, $sUnix95, $sSwitch, $iExitCode,
       %hsProc ) = ();

  $sSwitch = $^O;    # Get OS..
  @$rlProcList = (); # Truncate..

  #-------------
  # Store LANG..
  #-------------
  $sLang = $ENV { LANG };
  $ENV { LANG } = 'C';

  #---------------------------
  # Switch to the correct OS..
  #---------------------------
  if("solaris" eq $sSwitch) {

      $sPsCmd = "/usr/ucb/ps";
      $sPsCmdArgs = "axwwl $iPid";
      $sFmt = '[\da-z]+\s+([\d]+)\s+([\d]+)\s+([\d]+)\s+.*?[\d]:[\d]{2}\s(.*)';

  }

  elsif("hpux" eq $sSwitch) {

      #-----------------------
      # Force XPG4 behaviour..
      #-----------------------
      $sUnix95 = $ENV { UNIX95 };
      $ENV { UNIX95 } = 'XPG4';

      $sPsCmd = "/usr/bin/ps";
      $sPsCmdArgs = "-p $iPid -xxlo uid,pid,ppid,args";
      $sFmt = '([\d]+)\s+([\d]+)\s+([\d]+)\s+(.*)';

  }

  elsif("aix" eq $sSwitch) {

      $sPsCmd = "/usr/bin/ps";
      $sPsCmdArgs = "axwwl $iPid";
      $sFmt =  '[\da-e]+\s+[A-Z]\s+([\d]+)\s+([\d]+)\s+([\d]+)\s+.*?[\d]:[\d]{2}\s(.*)';

  }


  elsif("linux" eq $sSwitch) {

      $sPsCmd = "/bin/ps";
      $sPsCmdArgs = "wwo uid,pid,ppid,args $iPid";
      $sFmt = '([\d]+)\s+([\d]+)\s+([\d]+)\s+(.*)';

  }

  &logTrace ( 20152, "<%s> OS [%s] PsCmd [%s] PsCmdArgs [%s]",
  $sMyName, $sSwitch, $sPsCmd, $sPsCmdArgs );

  #-------------------
  # Validate command..
  #-------------------
  if ( ! -x $sPsCmd ) {
    &logError ( 20153,
    "<%s> ps(1b) command [%s] is not executable", $sMyName, $sPsCmd );
    return 0; # Failure..
  }

  #-------------------------------
  # Fire command and get details..
  #-------------------------------
  @hsProc { qw ( UID PID PPID COMMAND ) } =
  map {
    s/\n//;                             # Chomp..
    s/^\s+//;                           # Strip leading spaces..
    /^$sFmt$/o;                         # Extract fields..
  } readpipe ( "$sPsCmd $sPsCmdArgs" ); # From the command..

  $iExitCode = $? >> 8; # Get exit code..

  #------------------------
  # Restore env variables..
  #------------------------
  defined $sLang ? $ENV { LANG } = $sLang:delete $ENV { LANG };
  defined $sUnix95 ? $ENV { UNIX95 } = $sUnix95:delete $ENV { UNIX95 };

  #------------------
  # Check exit code..
  #------------------
  if ( $iExitCode ) {
    &logError ( 20154,
    "<%s> Cmd [%s] with args [%s] failed with exit code [%s]",
    $sMyName, $sPsCmd, $sPsCmdArgs, $iExitCode );

    %hsProc = (); # Truncate..
    return 0;
  }

  #---------------------
  # Fill up the bucket..
  #---------------------
  push ( @$rlProcList, \%hsProc ); # Update..

  return $iReturnCode;
}



=pod

 -------------------------------------------------------------------------
 Function : bool GetSrvrDetails ( IN String LCName,
                                  OUT RefHashList SrvrDetails );

 Description : Gives PID file and process string for given server type

 Parameters  :

 $sLCName                   Name of the liveCache instance.
 $rhlSrvrDetails            Stores details of PID file and process string
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub GetSrvrDetails {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sLCName,
       $rhlSrvrDetails ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace( 20155, "Subroutine <%s> called with:
  LCName                   [%s]
  SrvrDetails              [%s]",

  $sMyName, $sLCName, $rhlSrvrDetails );

  #-----------------------------
  # Set return code to success..
  #-----------------------------
  my $iReturnCode = 1; # Set to true..

  #-------------------
  # Define variables..
  #-------------------
  my ( $sKey, $sProc, $sPidFile, $sPIDDir, $sResourceType,
       $sResourceName, $iPat ) = ();

  #-------------------------------------
  # List of search patterns to monitor..
  #-------------------------------------
  my @lsSrchPatterns = ( "kernel $sLCName", 
                         "kernel $sLCName" );

  #----------------------------------------------
  # Derive our resource type and resource names..
  #----------------------------------------------
  $sResourceType = ( $0 =~ m@([^/]+)@g ) [-2];
  $sResourceName   = $ENV { VCS_LOG_RESOURCE_NAME };
  $sPIDDir = "/tmp/.VRTS$sResourceType/$sResourceName/PID";

  #------------------------------
  # Create PID folder if needed..
  #------------------------------
  if ( -d $sPIDDir ) {
    &logTrace ( 20156, "<%s> PID folder [%s] exists",
    $sMyName, $sPIDDir );
  }
  elsif ( ! mkpath ( $sPIDDir, 0, 0700 ) ) {
    &logWarn ( 20157, "<%s> Failed to create PID folder [%s]",
    $sMyName, $sPIDDir );
    return 0; # Cannot proceed..
  }

  %$rhlSrvrDetails = (); # Truncate..

  #-----------------------------------------------
  # Set appropriate filter strings and PID files..
  #-----------------------------------------------
  foreach $sProc ( @lsSrchPatterns ) {
    $sKey = 'srchstr_' . ++ $iPat;
    $sPidFile = "$sPIDDir/pidfile.$sKey";
    @{ $$rhlSrvrDetails { $sKey } } = ( $sPidFile, $sProc ); # Store..
  }

  return $iReturnCode;
}



=pod

 -------------------------------------------------------------------------
 Function : int GetEngineTimeout ( IN Bool UseMargin );

 Description : Calculates engine timeout for entry point
 
 Parameters  :

 $bUseMargin      Flag to indicate if safety margin is requested

                  boolean value in the set of {0,1} where
                  0 No safety margin needed
                  1 Safety margin to be used
 Returns:  Time in secs.
 -------------------------------------------------------------------------

=cut

sub GetEngineTimeout {
  #----------------
  # Get Arguments..
  #----------------
  my ( $bUseMargin ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace( 20158, "Subroutine <%s> called with:
  UseMargin  [%s]",
  $sMyName, $bUseMargin );

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $sAttribute, $sCommand, $sCommandArgs,
       $iEngineTimeout, $iExitCode ) = ();

  my ( $sEntryPointName, $sResourceType ) = reverse $0 =~ m@([^/]+)@g;
  $sResourceType = "SAPLiveCache" if $sResourceType !~ /\w/; # Ensure..

  #----------------------------------
  # Grace margins from VCS timeouts..
  #----------------------------------
  my %hiGraceMargin = ( 'MonitorTimeout', 10,   # Seconds..
                        'OnlineTimeout',  10,
                        'OfflineTimeout', 10,
                        'CleanTimeout',   20 );

  #-----------------------------------------
  # Form query string based on entry point..
  #-----------------------------------------
  $sCommand = $sCmdHatype; # Get global..
  ( $sAttribute = $sEntryPointName ) =~ s/(.*)/\u$1Timeout/;
  $sCommandArgs = "-value $sResourceType $sAttribute";

  #----------------------------------
  # Fire command and collect output..
  #----------------------------------
  ( $iEngineTimeout = readpipe ( "$sCommand $sCommandArgs" ) ) =~ s/\n//;

  #--------------------
  # Check exit status..
  #--------------------
  if ( $iExitCode = $? >> 8 ) {
    &logTrace ( 20159,
    "<%s> Command [%s] with Args [%s] failed with exitcode [%s]",
    $sMyName, $sCommand, $sCommandArgs, $iExitCode );
    $iEngineTimeout = $bUseMargin?$hiGraceMargin{$sAttribute}:0;
  }
  elsif ( ! $iEngineTimeout ) { # Empty value..
    &logTrace ( 20160, "<%s> Could not get [%s] from engine",
    $sMyName, $sAttribute );
    $iEngineTimeout = $bUseMargin?$hiGraceMargin{$sAttribute}:0;
  }
  else { # Success..
    &logTrace ( 20161, "<%s> Got [%s] as [%s] sec from engine",
    $sMyName, $sAttribute, $iEngineTimeout );
  }

  #----------------------------
  # Apply margin if requested.. 
  #----------------------------
  if ( $bUseMargin     &&                                 # Margin requested..
       $iEngineTimeout >= $hiGraceMargin{$sAttribute} ) { # And valid..
    $iEngineTimeout -= $hiGraceMargin{$sAttribute};       # Apply..

    &logTrace ( 20162,
    "<%s> Setting up [%s] sec margin. Returning [%s]",
    $sMyName, $hiGraceMargin{$sAttribute}, $iEngineTimeout );
  }

  return ( $iEngineTimeout ); # Secs..
}


=pod

 -------------------------------------------------------------------------
 Function : int IntentionalOffline ( IN Integer State,
                                     IN Integer IState,
                                     IN String  LCName,
                                     IN String  LCHome,
                                     IN String  DBMUserId,
                                     IN String  DBMPassword,
                                     IN String  ResourceName );

 Description : Handles the Resource faults from the intentional shutdown
               of the appplicatin (SAPLiveCache Server).  
 
 Parameters  :

 $iState          Resource current state. 
 $iIState         Resource internal/intermediate state.
 $sLCName         Name of the liveCahe Server.
 $sLCHome         Home directory for liveCache server binaries.
 $sDBMUserId      User name for DBM operator.
 $sDBMPassword    Password for DBM operator.
 $sResourceName   Name of the resource which needs to be protected from 
                  Intentional shutdowns.

 Returns : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 105  ONLINE with confidence level 50   VCSDef_MonitorExit::PARTIAL_ONLINE
 110  ONLINE with confidence level 100          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 -------------------------------------------------------------------------

=cut

sub IntentionalOffline {

  #----------------
  # Get Arguments..
  #----------------
  my ( $iState,
       $iIState,
       $sLCName,
       $sLCHome,
       $sDBMUserId,
       $sDBMPassword,
       $sResourceName ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace( 20163, "Subroutine <%s> called with:
  State             [%s],
  IState            [%s],
  LCName            [%s],
  LCHome            [%s],
  DBMUserId         [%s],
  DBMPassword       [%s],
  ResourceName      [%s]",

  $sMyName, $iState, $iIState, $sLCName, $sLCHome, 
  $sDBMUserId, $sDBMPassword, $sResourceName );

  #-----------------------------------
  # Declare all the variables needed..
  #-----------------------------------
  my $iFound     = VCSDef_MonitorExit::OFFLINE;
  my @lsLCStatus = (); 
  my $sSrchStr   = "stopping ok";

  #--------------------------------------------------
  # Determine the name of the lccluster status file..
  #--------------------------------------------------
  my $sClustFile = "/sapdb/data/wrk/$sLCName/symc_lccluster_$sLCName" ;

  &logTrace ( 20164,
  "<%s> Checking for the lccluster status file [%s].",
  $sMyName, $sClustFile );

  if ( ! -e $sClustFile ) {
    &logTrace ( 20165,
    "<%s> lccluster status file [%s] doesn't exist.",
    $sMyName, $sClustFile );

    return $iFound;
  }

  &logTrace ( 20166, "<%s> Looking for string [%s] from file contents.",
  $sMyName, $sSrchStr );

  if ( ! open( LCFILE, $sClustFile ) ) { # Open the file ..
    &logTrace ( 20167, "<%s> Failed to open the file [%s].",
    $sMyName, $sClustFile );

    return $iFound;
  }
  else {
    @lsLCStatus = <LCFILE>;
    close ( LCFILE );

    #--------------------------------------------
    # Attempt to locate search string in output..
    #--------------------------------------------
    if ( ! grep { /$sSrchStr/o } @lsLCStatus ) {
      &logTrace ( 20168,
      "<%s> Failed to locate search pattern [%s] in the file contents [%s]",
      $sMyName, $sSrchStr, join ( chr 012, @lsLCStatus ) );

      return $iFound;
    }
    else {
      if( VCSDef_ResourceState::ONLINE == $iState && 
          VCSDef_ResourceIState::NOT_WAITING == $iIState ) {

        my $sHasysOut = `/opt/VRTSvcs/bin/hasys -nodeid 2>&1`;
        chomp $sHasysOut;
        $sHasysOut =~ s/.*Node name: //;
        $sHasysOut =~ s/ Node ID: .*//;
        my $rsCurrentSystemName = $sHasysOut;

        &logTrace ( 20169,
        "<%s> hares offlining [%s] on [%s].",
        $sMyName, $sResourceName, $rsCurrentSystemName );

        system("/opt/VRTSvcs/bin/hares -offline $sResourceName -sys $rsCurrentSystemName");

        #----------------------------------------------
        # Remove the lccluster status file.. 
        #----------------------------------------------
        &logTrace ( 20170,
        "<%s> Removing the lccluster status file [%s].",  
        $sMyName, $sClustFile ); 
        unlink $sClustFile if ( -e $sClustFile ); # Remove LC status file..

        $iFound = VCSDef_MonitorExit::ONLINE;
        
        return $iFound;
      }
      else {
        &logTrace ( 20171,
        "<%s> Invalid states State [%s] IState [%s] to check for intentional offline.", 
        $sMyName, $iState, $iIState );

        return $iFound;
      }
    }
  } 

}


=pod

 -------------------------------------------------------------------------
 Function : bool RemoveIPCResourcesForSAPLiveCache ( IN String User,
                                                     IN String LCHome,
                                                     IN String LCName );

 Description : Removes the IPC resources for the liveCache instance.

 Parameters  :

 $sUser      User name for which IPC resources are removed.
 $sLCHome    Home directory for SAP liveCache instance.
 $sLCName    liveCache name for which IPC needs to be removed.

 Returns : boolean value in the set of {0,1} where
           0 Successfully removed IPCs for SAP liveCache instance.
           1 Failed to remove IPC resources for SAP liveCache instance.
 -------------------------------------------------------------------------

=cut

sub RemoveIPCResourcesForSAPLiveCache {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sUser,
       $sLCHome,
       $sLCName ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace( 20172, "Subroutine <%s> called with:
  User           [%s]
  LCHome         [%s]
  LCName         [%s]",
  $sMyName, $sUser, $sLCHome, $sLCName );

  #---------------------------------------------------------------------
  # Upper bound.. Time in EPOCH when entry pt started + VCS time limit..
  #---------------------------------------------------------------------
  my $iTillWhen = $^T + SAPLiveCache::GetEngineTimeout ( 1 );

  #-------------------------------------------------------------
  # Try to get independant data path for SAP liveCache Server..
  #-------------------------------------------------------------
  my $sCommand = "$sLCHome/dbmcli";
  my $sCommandArgs = " dbm_getpath indepdatapath";

  my ( @lsCmdout, $iExitCode, $iTimeout ) = ();
  my ( $sIndepDataPath, @liIPCshm, @liIPCsem, $sDbDir, $sUsDir, @lsDbFiles, @lsUsFiles ) = ();
  my $iReturnCode = 1;

  &logTrace ( 20173,
  "<%s> Getting independent data path for liveCache instance with Cmd [%s] Args [%s]",
  $sMyName, $sCommand, $sCommandArgs );

  #------------------
  # Request startup..
  #------------------
  if ( ( $iTimeout = $iTillWhen - time ) <= 10 ) { # Get time bearings..
    &logError ( 20174,
    "<%s> Insufficient time to get indepdatapath for liveCache instance. Short by [%s] sec",
    $sMyName, abs ( $iTimeout ) );
    $iReturnCode = 0; # Failure..
  }
  elsif ( $iReturnCode =
    SAPLiveCache::ExecLocalCmd ( "root",
                                 "",
                                 $sCommand,
                                 $sCommandArgs,
                                 \$iExitCode,
                                 $iTimeout,
                                 \@lsCmdout ) ) {
    &logTrace ( 20175,
    "<%s> Cmd [%s] with args [%s] executed successfully",
    $sMyName, $sCommand, $sCommandArgs );
  }
  else {
    &logError ( 20176,
    "<%s> Cmd [%s] with args [%s] failed with exit code [%s]. Cmd output [%s]",
    $sMyName, $sCommand, $sCommandArgs, $iExitCode,
    join ( chr 012, @lsCmdout ) );
    $iReturnCode = 1; 
  }

  if( scalar @lsCmdout ) {
    $sIndepDataPath = $lsCmdout[-1];
    chomp $sIndepDataPath;
  }
  else {
    $sIndepDataPath = "/sapdb/data" ;
    &logError(20177,
              "<%s> Failed to find the indepdatapath with the command [%s] with arguments [%s]",
              $sMyName, $sCommand, $sCommandArgs ); 
    &logInfo (20178,
              "<%s> Defaulting the indepdatapath for liveCache server [%s] to [%s]", 
              $sMyName, $sLCName, $sIndepDataPath ); 
  }

  $sDbDir = $sIndepDataPath . "/ipc/" . "db\:" . $sLCName;
  $sUsDir = $sIndepDataPath . "/ipc/" . "us\:" . $sLCName;

  if( ReadDirectory( $sDbDir,
                     \@lsDbFiles ) ) {  # Read from db:<LCName> directory ...
    &logTrace(20179,
              "<%s> Read files [%s] from directory [%s]",
              $sMyName, join(' ', @lsDbFiles), $sDbDir );
    if( scalar @lsDbFiles && SAPLiveCache::GetIPCIDsFromFiles ( \@lsDbFiles,
                                                                O_IPCSEM,
                                                                \@liIPCsem ) ) {
      &logTrace(20180,
                "<%s> Read semaphore IDs [%s] from the files in directory [%s]",
                $sMyName, join(' ', @liIPCsem), $sDbDir );
    }
    else {
      &logTrace(20181,
               "<%s> No semaphore IDs to read from the files in direcotory[%s]",
               $sMyName, $sDbDir );
    }

    if( scalar @lsDbFiles && SAPLiveCache::GetIPCIDsFromFiles ( \@lsDbFiles,
                                                                O_IPCSHM,
                                                                \@liIPCshm ) ) {
      &logTrace(20182,
                "<%s> Read Shared memory IDs [%s] from the files in directory [%s]",
                $sMyName, join(' ', @liIPCshm), $sDbDir );
    }
    else {
      &logTrace(20183,
                "<%s> No Shared Memory IDs to read from the files in direcotory[%s]",
                $sMyName, $sDbDir );
    }
  }
  else {
    &logTrace( 20184,
               "<%s> Skiping directory [%s] from reading for IPC.",
               $sMyName, $sDbDir );
  }

  if( SAPLiveCache::ReadDirectory( $sUsDir,
                                   \@lsUsFiles ) ) {  # Read from us:<LCName> directory ...
    &logTrace(20185,
              "<%s> Read files [%s] from directory [%s]",
              $sMyName, join(' ', @lsUsFiles), $sUsDir );
    if( scalar @lsUsFiles && SAPLiveCache::GetIPCIDsFromFiles ( \@lsUsFiles,
                                                                O_IPCSEM,
                                                                \@liIPCsem ) ) {
      &logTrace(20186,
                "<%s> Read semaphore IDs [%s] from the files in directory [%s]",
                $sMyName, join(' ', @liIPCsem), $sUsDir );
    }
    else {
      &logTrace(20187,
                "<%s> Failed to read to semaphore IDs from the files in direcotory[%s]",
                $sMyName, $sUsDir );
    }

    if( scalar @lsUsFiles && SAPLiveCache::GetIPCIDsFromFiles ( \@lsUsFiles,
                                                                O_IPCSHM,
                                                                \@liIPCshm ) ) {
      &logTrace(20188,
                "<%s> Read Shared memory IDs [%s] from the files in directory [%s]",
                $sMyName, join(' ', @liIPCshm), $sUsDir );
    }
    else {
      &logTrace(20189,
               "<%s> Failed to read to Shared Memory IDs from the files in direcotory[%s]",
               $sMyName, $sUsDir );
    }
  }
  else {
    &logTrace( 20190,
              "<%s> Failed to read directory [%s].",
              $sMyName, $sUsDir );
  }

  if( scalar @liIPCsem ) {  # Check Semaphores for removing ...
    if( $iReturnCode = SAPLiveCache::RemoveIPCResourcesForUser ( $sUser,
                                                                 O_IPCSEM,
                                                                 \@liIPCsem ) ) {
      &logTrace (20191,
                 "<%s> Successfully removed Semaphores IDs [%s] for User [%s]",
                 $sMyName, join(' ', @liIPCsem), $sUser );
    }
    else {
      &logError (20192,
                 "<%s> Failed to remove Semaphores for User [%s]",
                 $sMyName, $sUser );
      $iReturnCode = 0; # Failure ...
    }
  }

  if( scalar @liIPCshm ) {  # Check Shared Memory for removing ...
    if( $iReturnCode = SAPLiveCache::RemoveIPCResourcesForUser ( $sUser,
                                                                 O_IPCSHM,
                                                                 \@liIPCshm ) ) {
      &logTrace (20193,
                 "<%s> Successfully removed Shared Memory IDs [%s] for User [%s]",
                 $sMyName, join(' ', @liIPCshm), $sUser );
    }
    else {
      &logError (20194,
                 "<%s> Failed to remove Shared Memory IDs for User [%s]",
                 $sMyName, $sUser );

      $iReturnCode = 0; # Failure ...
    }
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool ReadDirectory ( IN String Dir,
                                 IN RefListOfString Files)

 Description : Reads all the files from the specified directory.

 Parameters  :

 $sDir      The directory from ehich the files needs to be read.
 $rlsFiles  The list of files read from the specified directory.

 Returns : boolean value in the set of {0,1} where
           0 Successfully read all the files from directory.
           1 Failed to read the files from directory.
 -------------------------------------------------------------------------

=cut

sub ReadDirectory {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sDir,
       $rlsFiles ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace( 20195, "Subroutine <%s> called with:
  Dir            [%s]
  rlsFiles       [%s]",
  $sMyName, $sDir, $rlsFiles );

  if( ! -d $sDir ) {
    &logTrace ( 20196,
    "<%s> Directory [%s] is not valid. Skipping ..",
    $sMyName, $sDir );

    return 0;  # Failure ...
  }

  my ( $sFile, $sPwd, $sCurr ) = ();

  if ( opendir ( DIRENT, $sDir ) ) {
    &logTrace( 20197,
    "<%s> Directory [%s] opened for reading.",
    $sMyName, $sDir );

    @$rlsFiles = readdir ( DIRENT );

    closedir ( DIRENT ) ||
    &logWarn(20198,
    "<%s> Failed to close directory [%s] after reading",
    $sMyName, $sDir );
  }
  else {
    &logWarn(20199,
    "<%s> Failed to open Directory [%s] for reading.",
    $sMyName, $sDir );

    return 0;  # Failure ...
  }

  return 1;
}


=pod

 -------------------------------------------------------------------------
 Function : bool GetIPCIDsFromFiles ( IN RefListOfString Files,
                                      IN Interger IPCFlags,
                                      OUT RefListOfInteger IPCs );

 Description : Returns all the IPC IDs from the IPC files.

 Parameters  :

 $rlsFiles      List of IPC file names from which the ID are to read.
 $riIPCFlags    The type of IPCs needs to be read, like O_IPCSEM, O_IPCSHM etc..
 $rliIPCs       List of IPC IDs read from the files.

 Returns : boolean value in the set of {0,1} where
           0 Successfully read IDs from files.
           1 Failed to read IDs from files.
 -------------------------------------------------------------------------

=cut

sub GetIPCIDsFromFiles {
  #----------------
  # Get Arguments..
  #----------------
  my ( $rlsFiles,
       $riIPCFlags,
       $rliIPCs ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace( 20200, "Subroutine <%s> called with:
  rlsFiles       [%s]
  IPC Flags      [%s]
  rliIPCs        [%s]",
  $sMyName, $rlsFiles, $riIPCFlags, $rliIPCs );

  foreach my $sFile (@$rlsFiles) {
    chomp $sFile;
    my $iIPC = $sFile;

    if( O_IPCSHM == $riIPCFlags && $iIPC =~ s/^-m//o ) {
      &logTrace(20201,
      "<%s> Got Shared Memory ID [%s] from [%s].",
      $sMyName, $iIPC, $sFile );

      push(@$rliIPCs, $iIPC);
    }
    elsif( O_IPCSEM == $riIPCFlags && $iIPC =~ s/^-s//o ) {
      &logTrace(20202,
      "<%s> Got Semaphore ID [%s] from [%s].",
      $sMyName, $iIPC, $sFile );

      push(@$rliIPCs, $iIPC);
    }
    elsif( O_IPCMSQ == $riIPCFlags && $iIPC =~ s/^-q//o ) {
      &logTrace(20203,
      "<%s> Got Message Queue ID [%s] from [%s].",
      $sMyName, $iIPC, $sFile );

      push(@$rliIPCs, $iIPC);
    }
    elsif( O_IPCALL == $riIPCFlags && $iIPC =~ s/^-[sqm]+//o ) {
      &logTrace(20204,
      "<%s> Got IPC ID [%s] from [%s].",
      $sMyName, $iIPC, $sFile );

      push(@$rliIPCs, $iIPC);
    }
  }

  return 1;
}

=pod

 -------------------------------------------------------------------------
 Function : bool RemoveIPCResourcesForUser ( IN String User,
                                             IN Integer IpcFlags,
                                             IN RefListInteger Ids );

 Description : Removes IPC resources for specified UNIX user

 Parameters :

 $sUser      UNIX user whose IPC resources are to be removed
 $iIpcFlags  Behaviour flag
 $rliIds     List of IPC IDs to remove for that user

 Values for Behaviour flag:
 O_IPCSHM : Shared Memory           O_IPCSEM : Semaphore
 O_IPCMSQ : Message Queue           O_IPCALL : All of the above

 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub RemoveIPCResourcesForUser {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sUser,
       $iIpcFlags,
       $rliIds ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace( 20205, "Subroutine <%s> called with:
  User           [%s]
  IPC Flags      [%s]
  rliIds         [%s]",
  $sMyName, $sUser, $iIpcFlags, $rliIds );

  #-------------------------
  # Define local variables..
  #-------------------------
  my $sFormat = 'l l l l ul ul i'; # Needed for unpack. Check intro(2)..
  my $iLocn = $^O eq 'linux'?1:0;  # Index locn. of ipc_perm.uid..
  my $lsVar = 0;                   # Stores the binary structure semid_ds..
  my $sCommand = $sCmdIpcs;        # Get global..

  my ( $iId, $iUid, $iIpcUid, $iRemoved,
       @liShmIds, @liMqIds, @liSemIds, @liIdsRemoved, %hiIds ) = ();

  #-------------------------------------
  # Set return code to hopeful success..
  #-------------------------------------
  my $iReturnCode = 1;

  #--------------------
  # Validate the user..
  #--------------------
  unless ( defined ( $iUid = getpwnam $sUser ) ) {
    &logError ( 20206, "<%s> Invalid User [%s] specfied. Cannot proceed.",
    $sMyName, $sUser );
    return 0; # Error..
  }

  #----------------
  # Check for IDs..
  #----------------
  if ( scalar @$rliIds ) {
    &logTrace ( 20207,
    "<%s> Request to remove only [%s] IDs for User [%s]",
    $sMyName, join ( ' ', @$rliIds ), $sUser );

    %hiIds = map { $_ => 1 } @$rliIds; # Remove duplicates..
  }
#  else {
#    &logWarn ( 20032,
#    "<%s> Called with [%s] IDs for User [%s]",
#    $sMyName, join ( ' ', @$rliIds ), $sUser );
#    return 1;
#  }

  #-------------------------------
  # Read in the contents of IPCs..
  #-------------------------------
  foreach ( readpipe ( $sCommand ) ) {
    s/\n//; # Chomp..

    #---------------
    # Handle Linux..
    #---------------
    if ( $^O eq 'linux' ) {
      if ( $iIpcFlags & O_IPCSHM && # Shared Memory..
           /shmid/ ... /^$/ ) {
        /(\w+)\s+(\d+)/;
        push ( @liShmIds, $2 ) if defined $2;
      }

      if ( $iIpcFlags & O_IPCSEM && # Semaphore.. 
           /semid/ ... /^$/ ) {
        /(\w+)\s+(\d+)/;
        push ( @liSemIds, $2 ) if defined $2;
      }

      if ( $iIpcFlags & O_IPCMSQ && # Message Queue.. 
           /msqid/ ... /^$/ ) {
        /(\w+)\s+(\d+)/;
        push ( @liMqIds, $2 ) if defined $2;
      }
    }
    else { # the rest..
      #--------------------
      # Skip removed IPCs..
      #--------------------
      if ( /^[m|s|q]\s*(\d+)\s+\w+\s+(\S)/ && # Of all three..
           defined $2                      && # Mode defined..
           ( $2 eq 'D' || $2 eq 'C' ) ) {     # Says, it's removed..
        next;                                 # Skip..
      }

      $iIpcFlags & O_IPCSHM && # Shared Memory..
      push ( @liShmIds, $_ =~ /^m\s*(\d+).*/ );

      $iIpcFlags & O_IPCMSQ && # Message Queue..
      push ( @liMqIds,  $_ =~ /^q\s*(\d+).*/ );

      $iIpcFlags & O_IPCSEM && # Semaphore..
      push ( @liSemIds, $_ =~ /^s\s*(\d+).*/ );
    }
  }

  #---------------------------------------------
  # Process each one separately as per request..
  #---------------------------------------------

  #----------------
  # Shared Memory..
  #----------------
  if ( $iIpcFlags & O_IPCSHM ) {
    foreach $iId ( @liShmIds ) {

      next if ( scalar @$rliIds && ! $hiIds { $iId } );

      shmctl( $iId, IPC_STAT, $lsVar ) ||
      &logError ( 20208,
      "<%s> Failed to use shmctl(2) on SHM Id [%s]", $sMyName, $iId );

      #-----------------------------------------------------
      # Unpack uid(2) out of the packed semid_ds structure..
      #-----------------------------------------------------
      $iIpcUid = ( unpack ( $sFormat, $lsVar ) ) [$iLocn];

      #-----------------------------
      # Remove resource if a match..
      #-----------------------------
      if ( $iIpcUid == $iUid ) {
        if ( shmctl ( $iId, IPC_RMID, 0 ) ) {
          push ( @liIdsRemoved, $iId );
          ++$iRemoved; # Increment..
        }
        else {
          &logError ( 20209,
          "<%s> Failed to remove SHM Id [%s]. Error [%s]", $sMyName, $iId, $! );
          $iReturnCode = 0; # Failed..
        }
      }
    }
  }

  if ( scalar @liIdsRemoved ) {
    &logTrace ( 20210, "<%s>
    +--- Removed SHM Ids: [%s] for User [%s] Uid [%s]. Total [%s] ---+",
    $sMyName, join ( ' ', @liIdsRemoved ), $sUser, $iUid, scalar @liIdsRemoved );
    @liIdsRemoved = (); # Clear..
  }

  #----------------
  # Message Queue..
  #----------------
  $lsVar = 0; # Reset..
  if ( $iIpcFlags & O_IPCMSQ ) {
    foreach $iId ( @liMqIds ) {

      next if ( scalar @$rliIds && ! $hiIds { $iId } );

      msgctl( $iId, IPC_STAT, $lsVar ) ||
      &logError ( 20211,
      "<%s> Failed to use msgctl(2) on SHM Id [%s]", $sMyName, $iId );

      #-----------------------------------------------------
      # Unpack uid(2) out of the packed semid_ds structure..
      #-----------------------------------------------------
      $iIpcUid = ( unpack ( $sFormat, $lsVar ) ) [$iLocn];

      #-----------------------------
      # Remove resource if a match..
      #-----------------------------
      if ( $iIpcUid == $iUid ) {
        if ( msgctl ( $iId, IPC_RMID, 0 ) ) {
          push ( @liIdsRemoved, $iId );
          ++$iRemoved; # Increment..
        }
        else {
          &logError ( 20212,
          "<%s> Failed to remove Msg QId [%s]. Error [%s]", $sMyName, $iId, $! );
          $iReturnCode = 0; # Failed..
        }
      }
    }
  }

  if ( scalar @liIdsRemoved ) {
    &logTrace ( 20213, "<%s>
    +--- Removed Msg QIds: [%s] for User [%s] Uid [%s]. Total [%s] ---+",
    $sMyName, join ( ' ', @liIdsRemoved ), $sUser, $iUid, scalar @liIdsRemoved );
    @liIdsRemoved = (); # Clear..
  }


  #------------
  # Semaphore..
  #------------
  $lsVar = 0; # Reset..
  if ( $iIpcFlags & O_IPCSEM ) {
    foreach $iId ( @liSemIds ) {

      next if ( scalar @$rliIds && ! $hiIds { $iId } );

      semctl( $iId, 0, IPC_STAT, $lsVar ) ||
      &logError ( 20214,
      "<%s> Failed to use semctl(2) on Sem Id [%s]", $sMyName, $iId );

      #-----------------------------------------------------
      # Unpack uid(2) out of the packed semid_ds structure..
      #-----------------------------------------------------
      $iIpcUid = ( unpack ( $sFormat, $lsVar ) ) [$iLocn];

      #-----------------------------
      # Remove resource if a match..
      #-----------------------------
      if ( $iIpcUid == $iUid ) {
        if ( semctl ( $iId, 0, IPC_RMID, 0 ) ) {
          push ( @liIdsRemoved, $iId );
          ++$iRemoved; # Increment..
        }
        else {
          &logError (20215,
          "<%s> Failed to remove Sem Id [%s]. Error [%s]", $sMyName, $iId, $! );
          $iReturnCode = 0; # Failed..
        }
      }
    }
  }

  if ( scalar @liIdsRemoved ) {
    &logTrace ( 20216,
    "<%s>
    +--- Removed Sem Ids: [%s] for User [%s] Uid [%s]. Total [%s] ---+",
    $sMyName, join ( ' ', @liIdsRemoved ), $sUser, $iUid, scalar @liIdsRemoved );
    @liIdsRemoved = (); # Clear..
  }

  #---------------------------------
  # Log, if we did anything useful..
  #---------------------------------
  if ( $iRemoved ) {
    &logTrace( 20217,
    "<%s> Successfully removed [%s] IPC resources for User [%s] in all.",
    $sMyName, $iRemoved, $sUser );
  }

  return $iReturnCode;
}

1; # So PERL knows its included..
